"""Minecraft Command Generator of Acacia."""

__all__ = ['Generator', 'Context']

from typing import TYPE_CHECKING, Union, Optional, List, Tuple, Dict
import contextlib

from acaciamc.ast import *
from acaciamc.error import *
from acaciamc.objects import *
from acaciamc.objects.none import ctdt_none
from acaciamc.constants import FUNCTION_PATH_CHARS
from acaciamc.mccmdgen.expr import *
from acaciamc.mccmdgen.symbol import SymbolTable, CTRTConversionError
from acaciamc.mccmdgen.mcselector import MCSelector
from acaciamc.mccmdgen.datatype import *
from acaciamc.mccmdgen.ctexecuter import CTExecuter
from acaciamc.mccmdgen.ctexpr import CTObj, CTObjPtr
from acaciamc.mccmdgen.utils import unreachable, InvalidOpError
from acaciamc.localization import localize
import acaciamc.mccmdgen.cmds as cmds

if TYPE_CHECKING:
    from acaciamc.compiler import Compiler
    from acaciamc.mccmdgen.datatype import DataType
    from acaciamc.mccmdgen.ctexpr import CTDataType, CTExpr

FUNC_NONE = "none"
FUNC_INLINE = "inline"
FUNC_NORMAL = "normal"

OP2METHOD = {
    Operator.add: 'add',
    Operator.minus: 'sub',
    Operator.multiply: 'mul',
    Operator.divide: 'div',
    Operator.mod: 'mod',
    Operator.positive: 'unarypos',
    Operator.negative: 'unaryneg',
    Operator.not_: 'unarynot'
}

class Context:
    def __init__(self, scope: SymbolTable):
        self.scope: SymbolTable = scope
        # current_function: the current function we are visiting
        self.current_function: \
            Optional[Union[AcaciaFunction, InlineFunction]] = None
        # inline_result: stores result expression of a function
        # only exists when visiting inline functions.
        self.inline_result: Optional[AcaciaExpr] = None
        # function_state: "none" | "inline" | "normal"
        # stores type of the nearest level of function. If not in any
        # function, it is "none".
        self.function_state: str = FUNC_NONE
        # self_value: value of `self` keyword
        self.self_value: Optional[AcaciaExpr] = None
        # entity_new_data: used by entity `new` system
        self.entity_new_data: Optional[Tuple[AcaciaExpr, str]] = None

    def copy(self):
        res = Context(self.scope)
        res.current_function = self.current_function
        res.inline_result = self.inline_result
        res.function_state = self.function_state
        res.self_value = self.self_value
        res.entity_new_data = self.entity_new_data
        return res

    def new_scope(self):
        self.scope = SymbolTable(self.scope, self.scope.builtins)

class Generator(ASTVisitor):
    """Generates MC function from an AST for a single file."""
    def __init__(self, node: AST, main_file: cmds.MCFunctionFile,
                 file_name: str, compiler: "Compiler"):
        super().__init__()
        self.node = node
        self.compiler = compiler
        self.file_name = file_name
        self.current_file = main_file
        self.ctx = Context(SymbolTable(builtins=compiler.builtins))
        # processing_node: prepared for showing errors
        # to know which AST we are passing (so that lineno and col are known)
        self.processing_node = self.node
        # node_depth: how deep we are in the tree (for debugging comments)
        self.node_depth: int = -1
        # current_tmp_scores: tmp scores allocated on current statement
        # see method `visit`.
        self.current_tmp_scores = []

    def parse(self):
        """Parse the AST and generate commands."""
        try:
            self.visit(self.node)
        except Error as err:
            self.error(err)

    def parse_as_module(self) -> AcaciaModule:
        """Parse the AST and return it as an `AcaciaModule`."""
        self.parse()
        return AcaciaModule(self.ctx.scope)

    def fix_error_location(self, error: Error):
        error.location.linecol = (
            self.processing_node.lineno,
            self.processing_node.col
        )

    # --- INTERNAL USE ---

    def lookup_symbol(self, name: str, **options) -> Optional[AcaciaExpr]:
        try:
            return self.ctx.scope.lookup(name, **options)
        except CTRTConversionError as err:
            self.error_c(ErrorType.NON_RT_NAME, name=name,
                         type_=err.expr.cdata_type.name)

    def attribute_of(self, primary: AcaciaExpr, attr: str):
        try:
            return primary.attribute_table.lookup(attr)
        except CTRTConversionError as err:
            self.error_c(ErrorType.NON_RT_ATTR,
                         primary=str(primary.data_type), attr=attr,
                         type_=err.expr.cdata_type.name)

    def error_c(self, *args, **kwds):
        self.error(Error(*args, **kwds))

    def error(self, error: Error):
        if not error.location.file_set():
            error.location.file = self.file_name
        if not error.location.linecol_set():
            self.fix_error_location(error)
        raise error

    def error_node(self, node: AST, *args, **kwds):
        err = Error(*args, **kwds)
        err.location.linecol = (node.lineno, node.col)
        self.error(err)

    def register_symbol(self, name: str, value: AcaciaExpr):
        v = self.lookup_symbol(name, use_outer=False, use_builtins=False)
        if v is not None:
            self.error_c(ErrorType.SHADOWED_NAME, name=name)
        self.ctx.scope.set(name, value)

    def node_location(self, node: AST) -> SourceLocation:
        return SourceLocation(self.file_name, (node.lineno, node.col))

    def write_debug(self, comment: str,
                    target: Optional[cmds.MCFunctionFile] = None):
        """Write debug comment to a file."""
        if target is None:
            target = self.current_file
        # write
        for comment_line in comment.split('\n'):
            target.write_debug(
                '# %s(%d:%d) %s' % (
                    ' ' * self.node_depth,
                    self.processing_node.lineno,
                    self.processing_node.col,
                    comment_line
                )
            )

    @contextlib.contextmanager
    def set_mcfunc_file(self, file: cmds.MCFunctionFile):
        old = self.current_file
        self.current_file = file
        yield
        self.current_file = old

    @contextlib.contextmanager
    def new_mcfunc_file(self, path: Optional[str] = None):
        """Create a new mcfunction file and set it to current file."""
        f = cmds.MCFunctionFile()
        with self.set_mcfunc_file(f):
            yield f
            if f.has_content():
                self.compiler.add_file(f, path)

    @contextlib.contextmanager
    def set_ctx(self, ctx: Context):
        old = self.ctx
        self.ctx = ctx
        yield
        self.ctx = old

    def new_ctx(self):
        return self.set_ctx(self.ctx.copy())

    # --- VISITORS ---

    def visit(self, node: AST, **kwargs):
        # store which node we are passing now
        old_node = self.processing_node
        self.processing_node = node
        self.node_depth += 1  # used by `self.write_debug`
        if isinstance(node, Statement):
            # NOTE `current_tmp_scores` is modified by `Compiler`, to
            # tell the tmp scores that are allocated in this statement
            # so that we can free them when the statement ends.
            # Therefore, only update tmp scores when node is a
            # `Statement`.
            old_tmp_scores = self.current_tmp_scores
            self.current_tmp_scores = []
        # write debug
        if node.show_debug:
            self.write_debug(type(node).__name__)
        # visit the node
        res = super().visit(node, **kwargs)
        # set back node info
        self.processing_node = old_node
        self.node_depth -= 1
        if isinstance(node, Statement):
            # free used vars
            for score in self.current_tmp_scores:
                self.compiler.free_tmp(score)
            self.current_tmp_scores = old_tmp_scores
        return res

    def visit_Module(self, node: Module):
        # start
        for stmt in node.body:
            self.visit(stmt)

    # --- STATEMENT VISITORS ---

    def visit_VarDef(self, node: VarDef):
        dt: DataType = self.visit(node.type)
        if not isinstance(dt, Storable):
            self.error_c(ErrorType.UNSUPPORTED_VAR_TYPE, var_type=str(dt))
        var = dt.new_var(self.compiler)
        self.register_symbol(node.target, var)
        if node.value is not None:
            self._assign(var, node.value)

    def visit_AutoVarDef(self, node: AutoVarDef):
        is_ctor: bool = False
        # Analyze data type & handle rvalue
        if isinstance(node.value, Call):
            func, table = self._call_inspect(node.value)
            value = self._call_invoke(node.value, func, table)
            is_ctor = isinstance(func, ConstructorFunction)
        else:
            value = self.visit(node.value)
        # Create variable
        dt = value.data_type
        if not isinstance(dt, Storable):
            self.error_c(ErrorType.UNSUPPORTED_VAR_TYPE, var_type=str(dt))
        if not is_ctor:
            var = dt.new_var(self.compiler)
        else:
            var = value
        self.register_symbol(node.target, var)
        # Assign
        if value is not None and not is_ctor:
            self.current_file.extend(value.export(var, self.compiler))

    def _assign(self, target: AcaciaExpr, value_node: Expression):
        if not target.is_assignable():
            self.error_c(ErrorType.INVALID_ASSIGN_TARGET)
        value_type: Optional[DataType] = None
        value: Optional[AcaciaExpr] = None
        # analyze rvalue
        if isinstance(value_node, Call):
            func, table = self._call_inspect(value_node)
            if isinstance(func, ConstructorFunction):
                value_type, ctor_kwds = \
                    func.pre_initialize(*table, self.compiler)
            else:
                value = self._call_invoke(value_node, func, table)
        else:
            value = self.visit(value_node)
        # check type
        if value_type is None:
            assert value is not None
            value_type = value.data_type
        if not target.data_type.matches(value_type):
            self.error_c(
                ErrorType.WRONG_ASSIGN_TYPE,
                expect=str(target.data_type), got=str(value_type)
            )
        # assign
        if value is not None:
            commands = value.export(target, self.compiler)
        else:
            commands = func.initialize(target, self.compiler, **ctor_kwds)
        # write commands
        self.current_file.extend(commands)

    def visit_Assign(self, node: Assign):
        target = self.visit(node.target)
        self._assign(target, node.value)

    def visit_AugmentedAssign(self, node: AugmentedAssign):
        method = f"i{OP2METHOD[node.operator]}"
        target = self.visit(node.target)
        value = self.visit(node.value)
        if not target.is_assignable():
            self.error_c(ErrorType.INVALID_ASSIGN_TARGET)
        # Call target's methods
        if hasattr(target, method):
            try:
                commands = getattr(target, method)(value, self.compiler)
            except InvalidOpError:
                pass
            else:
                # Success
                self.current_file.extend(commands)
                return
        # Error
        self._op_error_s(localize(f"{node.operator.value}.augmented"),
                         target, value)

    def visit_ReferenceDef(self, node: ReferenceDef):
        value: AcaciaExpr = self.visit(node.value)
        if node.type is not None:
            dt: DataType = self.visit(node.type)
            if not dt.matches(value.data_type):
                self.error_c(ErrorType.WRONG_REF_TYPE,
                             anno=str(dt), got=str(value.data_type))
        if not value.is_assignable():
            self.error_node(node.value, ErrorType.CANT_REF)
        self.register_symbol(node.name, value)

    def visit_ConstDef(self, node: ConstDef):
        for name, type_, value in zip(node.names, node.types, node.values):
            v = self.visit(value)
            if not isinstance(v, ConstExpr):
                self.error_node(value, ErrorType.NOT_CONST, name=name)
            if type_ is not None:
                dt: DataType = self.visit(type_)
                if not dt.matches(v.data_type):
                    self.error_node(
                        value,
                        ErrorType.WRONG_CONST_TYPE,
                        name=name, anno=str(dt), got=str(v.data_type)
                    )
            self.register_symbol(name, v)

    def visit_ExprStatement(self, node: ExprStatement):
        self.visit(node.value)

    def visit_Pass(self, node: Pass):
        pass

    def visit_FormattedStr(self, node: FormattedStr) -> str:
        res: List[str] = []
        for section in node.content:
            # expressions in commands need to be parsed
            if isinstance(section, str):
                res.append(section)
            else:
                expr: AcaciaExpr = self.visit(section)
                try:
                    value = expr.stringify()
                except InvalidOpError:
                    self.error_node(section, ErrorType.INVALID_FEXPR)
                res.append(value)
        return ''.join(res)

    def visit_Command(self, node: Command):
        command = cmds.Cmd(self.visit(node.content), suppress_special_cmd=True)
        self.current_file.write(command)

    def visit_If(self, node: If):
        # condition
        condition: AcaciaExpr = self.visit(node.condition)
        if not condition.data_type.matches_cls(BoolDataType):
            self.error_c(ErrorType.WRONG_IF_CONDITION,
                         got=str(condition.data_type))
        # process body
        with self.new_mcfunc_file() as body_file:
            self.write_debug('If body')
            for stmt in node.body:
                self.visit(stmt)
        has_then = body_file.has_content()
        with self.new_mcfunc_file() as else_body_file:
            self.write_debug('Else branch of If')
            for stmt in node.else_body:
                self.visit(stmt)
        has_else = else_body_file.has_content()
        if not has_then and not has_else:
            # no code in either branch
            return
        if has_then != has_else:
            new_cond = condition if has_then \
                else condition.unarynot(self.compiler)
            if isinstance(new_cond, SupportsAsExecute):
                # optimization for SupportsAsExecute (when only one of
                # then/else is present)
                dependencies, subcmds = new_cond.as_execute(self.compiler)
                self.current_file.extend(dependencies)
                self.current_file.write(cmds.Execute(
                    subcmds, runs=cmds.InvokeFunction(
                        body_file if has_then else else_body_file
                    )
                ))
                return
        # optimization: if condition is a constant, just run the code
        if isinstance(condition, BoolLiteral):
            self.current_file.extend(
                body_file.commands if condition.value
                else else_body_file.commands
            )
            return
        # Fallback
        dep, condition_var = to_BoolVar(condition, self.compiler, tmp=False)
        self.current_file.extend(dep)
        if has_then:
            self.current_file.write(cmds.Execute(
                [cmds.ExecuteScoreMatch(condition_var.slot, "1")],
                runs=cmds.InvokeFunction(body_file)
            ))
        if has_else:
            self.current_file.write(cmds.Execute(
                [cmds.ExecuteScoreMatch(condition_var.slot, "0")],
                runs=cmds.InvokeFunction(else_body_file)
            ))

    def visit_While(self, node: While):
        # condition
        condition: AcaciaExpr = self.visit(node.condition)
        if not condition.data_type.matches_cls(BoolDataType):
            self.error_c(ErrorType.WRONG_WHILE_CONDITION,
                         got=str(condition.data_type))
        # body
        with self.new_mcfunc_file() as body_file:
            self.write_debug('While definition')
            body_file.write_debug('## Part 1. Body')
            for stmt in node.body:
                self.visit(stmt)
        # process condition
        if isinstance(condition, BoolLiteral):
            # optimize when condition is known at compile time
            if condition.value is False:
                self.write_debug(
                    'Skipped because the condition always evaluates to False'
                )
                return
            else:
                self.error_c(ErrorType.ENDLESS_WHILE_LOOP)
        if isinstance(condition, SupportsAsExecute):
            # optimization for `SupportsAsExecute`
            dep, subcmds = condition.as_execute(self.compiler)
        else:
            dep, cond_var = to_BoolVar(condition, self.compiler, tmp=False)
            subcmds = [cmds.ExecuteScoreMatch(cond_var.slot, "1")]
        # triggering the function
        if body_file.has_content():
            body_file.write_debug('## Part 2. Recursion')
            for file in (body_file, self.current_file):
                file.extend(dep)
                file.write(cmds.Execute(
                    subcmds, runs=cmds.InvokeFunction(body_file)
                ))
        else:
            self.write_debug('No commands generated')

    def visit_InterfaceDef(self, node: InterfaceDef):
        if isinstance(node.path, str):
            path = node.path
        else:
            s: String = self.visit(node.path)
            path = s.value
        # Make sure the path is valid
        if not path:
            self.error_c(ErrorType.INTERFACE_PATH_EMPTY)
        if path.startswith('/'):
            self.error_c(ErrorType.INTERFACE_PATH_SLASH_START)
        if path.endswith('/'):
            self.error_c(ErrorType.INTERFACE_PATH_SLASH_END)
        if '//' in path:
            self.error_c(ErrorType.INTERFACE_PATH_DOUBLE_SLASH)
        for c in path:
            if c not in FUNCTION_PATH_CHARS:
                self.error_c(ErrorType.INTERFACE_PATH_INVALID_CHAR, char=c)
        # Make sure the path is unique and not reserved
        if self.compiler.is_reserved_path(path):
            self.error_c(ErrorType.RESERVED_INTERFACE_PATH, path=path)
        location = self.compiler.lookup_interface(path)
        if location is not None:
            err = Error(ErrorType.DUPLICATE_INTERFACE, path=path)
            err.add_frame(ErrFrame(location, "First occurrence", note=None))
            self.error(err)
        self.compiler.add_interface(path, self.node_location(node))
        with self.new_ctx():
            self.ctx.new_scope()
            # body
            with self.new_mcfunc_file(path) as body_file:
                self.write_debug('Interface definition')
                for stmt in node.body:
                    self.visit(stmt)
            # add lib
            if body_file.has_content():
                self.write_debug('Generated at %s' % body_file.get_path())
            else:
                self.write_debug('No commands generated')

    def visit_FunctionPort(self, node: FunctionPort):
        return (
            None if node.type is None else self.visit(node.type),
            node.port
        )

    def visit_ArgumentTable(self, node: ArgumentTable):
        # handle arg table
        args = node.args
        types: Dict[str, Optional[DataType]] = dict.fromkeys(args)
        defaults: Dict[str, Optional[AcaciaExpr]] = dict.fromkeys(args)
        ports: Dict[str, FuncPortType] = {}
        for arg in args:
            default_node = node.default[arg]
            type_node = node.types[arg]
            if default_node is not None:
                defaults[arg] = self.visit(default_node)
            types[arg], ports[arg] = self.visit(type_node)
            # make sure default value matches type
            # e.g. `def f(a: int = True)`
            if (defaults[arg] is not None
                and types[arg] is not None
                and not types[arg].is_type_of(defaults[arg])):
                self.error_c(
                    ErrorType.UNMATCHED_ARG_DEFAULT_TYPE,
                    arg=arg, arg_type=str(types[arg]),
                    default_type=str(defaults[arg].data_type)
                )
        return args, types, defaults, ports

    def visit_TypeSpec(self, node: TypeSpec):
        type_ = self.visit(node.content)
        try:
            dt = type_.datatype_hook()
        except InvalidOpError:
            self.error_c(ErrorType.INVALID_TYPE_SPEC, got=str(type_.data_type))
        return dt

    def visit_CallTable(self, node: CallTable):
        args: List[AcaciaExpr] = []
        keywords: Dict[str, AcaciaExpr] = {}
        for value in node.args:
            args.append(self.visit(value))
        for arg, value in node.keywords.items():
            keywords[arg] = self.visit(value)
        return args, keywords

    def _func_expr(self, fname: str, node: NormalFuncData) -> AcaciaFunction:
        """Return the function object to a function definition
        without parsing the body.
        """
        # get return type (of DataType type)
        if node.returns is None:
            returns = NoneDataType()
        else:
            returns = self.visit(node.returns)
        # check result type
        if not isinstance(returns, Storable):
            self.error_c(ErrorType.UNSUPPORTED_RESULT_TYPE,
                         result_type=str(returns))
        # parse arg
        args, types, defaults, _ = self.visit(node.arg_table)
        # infer missing type specs
        for name in args:
            if types[name] is None:
                types[name] = defaults[name].data_type
        # make sure default values are constants
            if defaults[name] is None:
                continue
            if not isinstance(defaults[name], ConstExpr):
                self.error_node(
                    node.arg_table.default[name],
                    ErrorType.NONREF_ARG_DEFAULT_NOT_CONST, arg=name
                )
        # check argument type
        for name, value in types.items():
            if not isinstance(value, Storable):
                self.error_c(ErrorType.UNSUPPORTED_ARG_TYPE,
                             arg=name, arg_type=str(value))
        # create function
        return AcaciaFunction(
            name=fname, args=args, arg_types=types, arg_defaults=defaults,
            returns=returns, compiler=self.compiler,
            source=self.node_location(node)
        )

    @contextlib.contextmanager
    def _in_noninline_func(self, func: AcaciaFunction):
        with self.new_ctx():
            self.ctx.new_scope()
            self.ctx.current_function = func
            self.ctx.function_state = FUNC_NORMAL
            yield

    def visit_InlineFuncData(self, node: InlineFuncData, name: str) \
            -> InlineFunction:
        # Return the inline function object to a function definition
        if node.returns is None:
            returns = NoneDataType()
            res_port = FuncPortType.by_value
        else:
            returns, res_port = self.visit(node.returns)
        args, arg_types, arg_defaults, arg_ports = self.visit(node.arg_table)
        for arg, default in arg_defaults.items():
            if default is None:
                continue
            if arg_ports[arg] is FuncPortType.by_reference:
                if not default.is_assignable():
                    self.error_node(node.arg_table.default[arg],
                                    ErrorType.CANT_REF_ARG, arg=arg)
            else:
                if not isinstance(default, ConstExpr):
                    self.error_node(
                        node.arg_table.default[arg],
                        ErrorType.NONREF_ARG_DEFAULT_NOT_CONST, arg=arg
                    )
        return InlineFunction(
            name, node,
            args, arg_types, arg_defaults, arg_ports,
            returns, res_port, self.ctx.copy(), owner=self,
            source=self.node_location(node)
        )

    def visit_NormalFuncData(self, node: NormalFuncData, name: str) \
            -> AcaciaFunction:
        func = self._func_expr(name, node)
        with self._in_noninline_func(func), \
             self.new_mcfunc_file() as body_file:
            # Register arguments to scope
            for arg, var in func.arg_vars.items():
                self.register_symbol(arg, var)
            # Write file
            self.write_debug('Function definition of %s()' % name)
            for stmt in node.body:
                self.visit(stmt)
        # Add file
        if body_file.has_content():
            self.write_debug('Generated at %s' % body_file.get_path())
            func.file = body_file
        else:
            self.write_debug('No commands generated')
        return func

    def visit_ConstFuncData(self, node: ConstFuncData, name: str) \
            -> AcaciaCTFunction:
        args = node.arg_table.args
        arg_types: Dict[str, Optional["CTDataType"]] = dict.fromkeys(args)
        defaults: Dict[str, Optional[CTObj]] = dict.fromkeys(args)
        ctexec = CTExecuter(self.ctx.scope, self, self.file_name)
        for arg in args:
            default_node = node.arg_table.default[arg]
            port_node = node.arg_table.types[arg]
            if default_node is not None:
                defaults[arg] = ctexec.visittop(default_node)
            if port_node is not None and port_node.type is not None:
                arg_types[arg] = ctexec.visittop(port_node.type)
            # make sure default value matches type
            # e.g. `def f(a: int = True)`
            if (
                defaults[arg] is not None
                and arg_types[arg] is not None
                and not arg_types[arg].is_typeof(defaults[arg])
            ):
                self.error_c(
                    ErrorType.UNMATCHED_ARG_DEFAULT_TYPE,
                    arg=arg, arg_type=arg_types[arg].name,
                    default_type=defaults[arg].cdata_type.name
                )
        if node.returns is None or node.returns.type is None:
            returns = ctdt_none
        else:
            returns = ctexec.visittop(node.returns.type)
        return AcaciaCTFunction(
            name, node, args, arg_types, defaults,
            returns, self.ctx, owner=self,
            source=self.node_location(node)
        )

    def visit_FuncDef(self, node: FuncDef):
        func: AcaciaCallable = self.visit(node.data, name=node.name)
        func.source = self.node_location(node)
        self.register_symbol(node.name, func)

    def module_traced(self, meta: ModuleMeta, lineno: int, col: int):
        try:
            res = self.compiler.parse_module(meta, self.current_file)
        except Error as err:
            err.add_frame(ErrFrame(
                SourceLocation(self.file_name, (lineno, col)),
                "Importing %s" % str(meta),
                note=None
            ))
            raise
        return res

    def visit_Import(self, node: Import):
        module, path = self.module_traced(node.meta, node.lineno, node.col)
        self.write_debug("Got module from %s" % path)
        self.register_symbol(node.name, module)
        self.ctx.scope.no_export.add(node.name)

    def visit_FromImport(self, node: FromImport):
        module, path = self.module_traced(node.meta, node.lineno, node.col)
        self.write_debug("Import from %s" % path)
        for name, alias in node.id2name.items():
            value = self.attribute_of(module, name)
            if value is None:
                self.error_c(ErrorType.MODULE_NO_ATTRIBUTE,
                             attr=name, module=str(node.meta))
            self.register_symbol(alias, value)
        self.ctx.scope.no_export.update(node.id2name.values())

    def visit_FromImportAll(self, node: FromImportAll):
        module, path = self.module_traced(node.meta, node.lineno, node.col)
        self.write_debug("Import everything from %s" % path)
        for name in module.attribute_table.all_names():
            self.register_symbol(name, self.attribute_of(module, name))
            self.ctx.scope.no_export.add(name)

    def _method_body(
        self, method: AcaciaFunction, self_var: TaggedEntity,
        template_name: str, body: List[Statement]
    ):
        with self._in_noninline_func(method):
            # Set self
            self.ctx.self_value = self_var
            # Register arguments to scope
            for arg, var in method.arg_vars.items():
                self.register_symbol(arg, var)
            # Write commands
            with self.set_mcfunc_file(method.file):
                self.write_debug('Entity method definition of %s.%s()' %
                                 (template_name, method.name))
                for stmt in body:
                    self.visit(stmt)

    def visit_EntityTemplateDef(self, node: EntityTemplateDef):
        field_types = {}  # Field name to `DataType`
        field_metas = {}  # Field name to field meta
        methods = {}  # Method name to function `AcaciaExpr`
        method_qualifiers = {}  # Method name to qualifier
        method_new = (None if node.new_method is None
                      else self.visit(node.new_method))
        # Handle parents
        parents = []
        for parent_ast in node.parents:
            parent = self.visit(parent_ast)
            if not parent.data_type.matches_cls(ETemplateDataType):
                self.error_c(ErrorType.INVALID_ETEMPLATE,
                             got=str(parent.data_type))
            parents.append(parent)
        # If parent is not specified, use builtin `Entity`
        if not parents:
            parents.append(self.compiler.base_template)
        # 1st Pass: get all the attributes and give every non-inline
        # method a `MCFunctionFile` without parsing its body.
        methods_2ndpass: List[Tuple[AcaciaFunction, FuncDef]] = []
        names_seen = set()
        def _name(node: AST, name: str):
            if name in names_seen:
                self.error_node(node, ErrorType.DUPLICATE_EFIELD, name=name)
            names_seen.add(name)
        for decl in node.body:
            res = self.visit(decl)
            if isinstance(decl, EntityField):
                # `res` is (field type, field meta)
                _name(decl, decl.name)
                field_types[decl.name], field_metas[decl.name] = res
            elif isinstance(decl, EntityMethod):
                # `res` is `AcaciaFunction`, `InlineFunction` or
                # `AcaciaCTFunction`
                _name(decl, decl.content.name)
                methods[decl.content.name] = res
                method_qualifiers[decl.content.name] = decl.qualifier
                if (
                    isinstance(res, AcaciaFunction)
                    and decl.qualifier is not MethodQualifier.static
                ):
                    methods_2ndpass.append((res, decl.content))
            else:
                assert isinstance(decl, Pass)
        # generate the template before 2nd pass, since `self` value
        # needs the template specified.
        template = EntityTemplate(
            node.name, field_types, field_metas,
            methods, method_qualifiers, method_new, parents,
            self.compiler, source=self.node_location(node)
        )
        # 2nd Pass: parse body of non-inline non-static methods.
        for method, ast in methods_2ndpass:
            if ast.name in template.method_dispatchers:
                disp = template.method_dispatchers[ast.name]
                _, gself = disp.impls[method]
                self_var = gself()
            else:
                assert ast.name in template.simple_methods
                self_var = template.simple_methods[ast.name].get_self_var()
            assert self_var is not None
            data = ast.data
            assert isinstance(data, NormalFuncData)
            # The self var getters will make sure the returned value is
            # not assignable, so no need to wrap it in `EntityReference`
            # here.
            self._method_body(method, self_var, node.name, data.body)
        # Also 2nd pass the `new` method if needed
        if isinstance(method_new, AcaciaNewFunction):
            nd_slot = method_new.template_id_var
            nd_tag = method_new.self_tag
            f = method_new.impl.file
            f.write_debug("## Clear tag used by `new` methods")
            f.write(f"tag @e[tag={nd_tag}] remove {nd_tag}")
            with self.new_ctx():
                self.ctx.entity_new_data = (IntVar(nd_slot), nd_tag)
                self_var = TaggedEntity(nd_tag, template)
                self_var.is_temporary = True
                self._method_body(
                    method_new.impl, self_var,
                    node.name, node.new_method.data.body
                )
        # Register
        self.register_symbol(node.name, template)

    def visit_EntityField(self, node: EntityField):
        data_type = self.visit(node.type)
        if not isinstance(data_type, SupportsEntityField):
            self.error_c(ErrorType.UNSUPPORTED_EFIELD_TYPE,
                         field_type=str(data_type))
        field_meta = data_type.new_entity_field(self.compiler)
        return data_type, field_meta

    def visit_EntityMethod(self, node: EntityMethod):
        data = node.content.data
        name = node.content.name
        if (isinstance(data, NormalFuncData)
                and node.qualifier is not MethodQualifier.static):
            func = self._func_expr(name, data)
            # Give this function a file
            file = cmds.MCFunctionFile()
            func.file = file
            self.compiler.add_file(file)
        else:
            func = self.visit(data, name=name)
        func.source = self.node_location(node)
        return func

    def visit_NewMethod(self, node: NewMethod):
        d = node.data
        inline = isinstance(d, InlineFuncData)
        if inline:
            func = self.visit(d, name="<new>")
        else:
            func = self._func_expr("<new>", d)
            if not func.result_type.matches_cls(NoneDataType):
                self.error_c(ErrorType.ENTITY_NEW_RETURN_TYPE,
                             got=func.result_type)
            # Give this function a file
            file = cmds.MCFunctionFile()
            func.file = file
            self.compiler.add_file(file)
        func.source = self.node_location(node)
        if not inline:
            # Allocate data needed by `new` system
            nd_slot = self.compiler.allocate()
            nd_tag = self.compiler.allocate_entity_tag()
            func = AcaciaNewFunction(func, nd_tag, nd_slot)
        return func

    def visit_For(self, node: For):
        iterable = self.visit(node.expr)
        # for-in on an entity group has completely different meaning --
        # it is not a compile-time loop.
        if not iterable.data_type.matches_cls(EGroupDataType):
            try:
                items = iterable.iterate()
            except InvalidOpError:
                self.error_c(ErrorType.NOT_ITERABLE,
                             type_=str(iterable.data_type))
            self.write_debug("Iterating over %d items" % len(items))
            for value in items:
                with self.new_ctx():
                    self.ctx.new_scope()
                    self.register_symbol(node.name, value)
                    for stmt in node.body:
                        self.visit(stmt)
        else:
            assert isinstance(iterable, EntityGroup)
            with self.new_mcfunc_file() as body_file, \
                 self.new_ctx():
                self.ctx.new_scope()
                self.write_debug("Entity group iteration body")
                this = TaggedEntity.new_tag(iterable.template, self.compiler)
                executer = EntityReference(MCSelector("s"), iterable.template)
                self.current_file.extend(executer.export(this, self.compiler))
                self.register_symbol(node.name, this)
                for stmt in node.body:
                    self.visit(stmt)
            self.write_debug("Entity group iteration at %s"
                             % body_file.get_path())
            self.current_file.write(cmds.Execute(
                [cmds.ExecuteEnv("as", iterable.get_selector().to_str())],
                runs=cmds.InvokeFunction(body_file)
            ))

    def visit_StructField(self, node: StructField):
        # Check whether type is storable.
        data_type = self.visit(node.type)
        if not isinstance(data_type, Storable):
            self.error_c(ErrorType.UNSUPPORTED_SFIELD_TYPE,
                         field_type=str(data_type))
        return node.name, data_type

    def visit_StructDef(self, node: StructDef):
        base_structs = list(map(self.visit, node.bases))
        for i, base in enumerate(base_structs):
            if not isinstance(base, StructTemplate):
                self.error_node(node.bases[i], ErrorType.INVALID_STEMPLATE,
                                got=str(base.data_type))
        fields: Dict[str, "DataType"] = {}
        for decl in node.body:
            res = self.visit(decl)
            if isinstance(decl, StructField):
                name, type_ = res
                if name in fields:
                    self.error_c(ErrorType.SFIELD_MULTIPLE_DEFS, name=name)
                fields[name] = type_
            else:
                assert isinstance(decl, Pass)
        self.register_symbol(node.name, StructTemplate(
            node.name, fields, base_structs,
            source=self.node_location(node)
        ))

    def visit_Result(self, node: Result):
        rt = self.ctx.current_function.result_type
        def _check(t: "DataType"):
            if rt is not None and not rt.matches(t):
                self.error_c(ErrorType.WRONG_RESULT_TYPE,
                             expect=str(rt), got=str(t))
        if self.ctx.function_state == FUNC_NORMAL:
            if isinstance(node.value, Call):
                func, table = self._call_inspect(node.value)
                if isinstance(func, ConstructorFunction):
                    value = None
                else:
                    value = self._call_invoke(node.value, func, table)
            else:
                value = self.visit(node.value)
            rv = self.ctx.current_function.result_var
            if value is None:
                dt, kwds = func.pre_initialize(*table, self.compiler)
                _check(dt)
                commands = func.initialize(rv, self.compiler, **kwds)
            else:
                _check(value.data_type)
                commands = value.export(rv, self.compiler)
            self.current_file.extend(commands)
        elif self.ctx.function_state == FUNC_INLINE:
            value: AcaciaExpr = self.visit(node.value)
            _check(value.data_type)
            p = self.ctx.current_function.result_port
            if p is FuncPortType.by_value:
                dt = value.data_type
                if not isinstance(dt, Storable):
                    self.error_c(ErrorType.UNSUPPORTED_RESULT_TYPE,
                                 result_type=str(dt))
                if self.ctx.inline_result is None:
                    self.ctx.inline_result = dt.new_var(self.compiler)
                    self.ctx.inline_result.is_temporary = True
                self.current_file.extend(
                    value.export(self.ctx.inline_result, self.compiler)
                )
            elif p is FuncPortType.by_reference:
                if self.ctx.inline_result is not None:
                    self.error_c(ErrorType.MULTIPLE_RESULTS)
                if not value.is_assignable():
                    self.error_c(ErrorType.CANT_REF_RESULT)
                self.ctx.inline_result = value
            else:
                assert p is FuncPortType.const
                if self.ctx.inline_result is not None:
                    self.error_c(ErrorType.MULTIPLE_RESULTS)
                if not isinstance(value, ConstExpr):
                    self.error_c(ErrorType.RESULT_NOT_CONST)
                self.ctx.inline_result = value
        else:
            assert self.ctx.function_state == FUNC_NONE
            self.error_c(ErrorType.RESULT_OUT_OF_SCOPE)

    def visit_NewCall(self, node: NewCall):
        if self.ctx.entity_new_data is None:
            self.error_c(ErrorType.NEW_OUT_OF_SCOPE)
        if node.primary is None:
            primary = self.compiler.base_template
        else:
            primary: AcaciaExpr = self.visit(node.primary)
            if not isinstance(primary, EntityTemplate):
                self.error_c(ErrorType.INVALID_ETEMPLATE,
                             got=primary.data_type)
        args, keywords = self.visit(node.call_table)
        try:
            commands = primary.method_new(
                self.compiler, *self.ctx.entity_new_data, args, keywords
            )
        except Error as err:
            s = primary.method_new_source
            err.add_frame(ErrFrame(
                self.node_location(node),
                localize("generator.visit.newcall.tracemsg") % primary.name,
                note = None if s is None else
                    localize("generator.visit.newcall.tracenote") % s
            ))
            raise
        self.current_file.extend(commands)

    # --- EXPRESSION VISITORS ---
    # literal

    def visit_Literal(self, node: Literal):
        value = node.value
        # NOTE Python bool is a subclass of int!!!
        if isinstance(value, bool):
            return BoolLiteral(value)
        elif isinstance(value, int):
            return IntLiteral(value)
        elif value is None:
            r = NoneLiteral()
            r.is_temporary = True
            return r
        elif isinstance(value, float):
            return Float(value)
        unreachable()

    def visit_StrLiteral(self, node: StrLiteral):
        return String(self.visit(node.content))

    def visit_Self(self, node: Self):
        v = self.ctx.self_value
        if v is None:
            self.error_c(ErrorType.SELF_OUT_OF_SCOPE)
        return v

    def _ct_list(self, items: List[Expression]):
        res = []
        for item in items:
            r = self.visit(item)
            if not isinstance(r, ConstExpr):
                self.error_node(item, ErrorType.ELEMENT_NOT_CONST)
            res.append(r)
        return res

    def visit_ListDef(self, node: ListDef):
        return AcaciaList(self._ct_list(node.items))

    def visit_MapDef(self, node: MapDef):
        return Map(self._ct_list(node.keys),
                   self._ct_list(node.values))

    # assignable

    def visit_Identifier(self, node: Identifier):
        res = self.lookup_symbol(node.name)
        if res is None:
            self.error_c(ErrorType.NAME_NOT_DEFINED, name=node.name)
        return res

    def visit_Attribute(self, node: Attribute):
        value = self.visit(node.object)
        res = self.attribute_of(value, node.attr)
        if res is None:
            self.error_c(
                ErrorType.HAS_NO_ATTRIBUTE,
                value_type=str(value.data_type), attr=node.attr,
            )
        return res

    # operators

    def _op_error_s(self, operator: str, *operands: AcaciaExpr):
        self.error_c(
            ErrorType.INVALID_OPERAND,
            operator=operator,
            operand=", ".join(
                '"%s"' % operand.data_type
                for operand in operands
            )
        )

    def _op_error(self, operator: Operator, *operands: AcaciaExpr):
        self._op_error_s(operator.localized, *operands)

    def visit_UnaryOp(self, node: UnaryOp):
        operand = self.visit(node.operand)
        try:
            res = getattr(operand, OP2METHOD[node.operator])(self.compiler)
        except InvalidOpError:
            self._op_error(node.operator, operand)
        return res

    def visit_BinOp(self, node: BinOp):
        left, right = self.visit(node.left), self.visit(node.right)
        meth = OP2METHOD[node.operator]
        try:
            res = getattr(left, meth)(right, self.compiler)
        except InvalidOpError:
            try:
                res = getattr(right, f"r{meth}")(left, self.compiler)
            except InvalidOpError:
                self._op_error(node.operator, left, right)
        return res

    def visit_CompareOp(self, node: CompareOp):
        compares = []
        left, right = None, self.visit(node.left)
        # split `e0 o1 e1 o2 e2 ... o(n) e(n)` into
        # `e0 o1 e1 and ... and e(n-1) o(n) e(n)`
        for operand, operator in zip(node.operands, node.operators):
            left, right = right, self.visit(operand)
            try:
                res = left.compare(operator, right, self.compiler)
            except InvalidOpError:
                try:
                    res = right.compare(
                        COMPOP_SWAP[operator], left, self.compiler
                    )
                except InvalidOpError:
                    self.error_c(
                        ErrorType.INVALID_OPERAND,
                        operator=operator.localized,
                        operand=f'"{left.data_type}", "{right.data_type}"'
                    )
            compares.append(res)
        return new_and_group(compares, self.compiler)

    def visit_BoolOp(self, node: BoolOp):
        operands = [self.visit(operand) for operand in node.operands]
        operator = node.operator
        # Make sure operands are all boolean
        for i, operand in enumerate(operands):
            if not operand.data_type.matches_cls(BoolDataType):
                self.error_node(
                    node.operands[i],
                    ErrorType.INVALID_BOOLOP_OPERAND,
                    operator=operator.localized,
                    operand=str(operand.data_type)
                )
        # Go
        if operator is Operator.and_:
            return new_and_group(operands, self.compiler)
        elif operator is Operator.or_:
            return new_or_expression(operands, self.compiler)
        unreachable()

    # call

    def _call_inspect(self, node: Call) -> \
            Tuple[AcaciaCallable, Tuple[ARGS_T, KEYWORDS_T]]:
        func = self.visit(node.func)
        if not isinstance(func, AcaciaCallable):
            self.error_c(ErrorType.UNCALLABLE, expr_type=str(func.data_type))
        table = self.visit(node.table)
        return func, table

    def _call_invoke(self, node: Call, func: AcaciaCallable,
                     table: Tuple[ARGS_T, KEYWORDS_T]) -> AcaciaExpr:
        # call it
        res, commands = func.call_withframe(
            *table, self.compiler,
            location=self.node_location(node)
        )
        # write commands
        self.current_file.extend(commands)
        return res

    def visit_Call(self, node: Call):
        func, table = self._call_inspect(node)
        return self._call_invoke(node, func, table)

    def call_function(self, func: AcaciaExpr, args: ARGS_T,
                      keywords: KEYWORDS_T, location=None) -> AcaciaExpr:
        if not isinstance(func, AcaciaCallable):
            self.error_c(ErrorType.UNCALLABLE, expr_type=str(func.data_type))
        res, commands = func.call_withframe(
            args, keywords, self.compiler, location
        )
        self.current_file.extend(commands)
        return res

    def call_inline_func(self, func: InlineFunction,
                         args: ARGS_T, keywords: KEYWORDS_T) -> CALLRET_T:
        # We visit the AST node every time an inline function is called
        file = cmds.MCFunctionFile()
        with self.set_ctx(func.context), self.new_ctx(), \
             self.set_mcfunc_file(file):
            self.ctx.new_scope()
            self.ctx.inline_result = None
            self.ctx.current_function = func
            self.ctx.function_state = FUNC_INLINE
            # Register args into scope
            arg2value = func.arg_handler.match(args, keywords)
            for arg, value in arg2value.items():
                port = func.arg_ports[arg]
                if port is FuncPortType.by_reference:
                    if not value.is_assignable():
                        self.error_c(ErrorType.CANT_REF_ARG, arg=arg)
                    self.register_symbol(arg, value)
                elif port is FuncPortType.by_value:
                    dt = value.data_type
                    if not isinstance(dt, Storable):
                        self.error_c(ErrorType.UNSUPPORTED_ARG_TYPE,
                                     arg=arg, arg_type=str(dt))
                    tmp = dt.new_var(self.compiler)
                    self.register_symbol(arg, tmp)
                    self.current_file.extend(value.export(tmp, self.compiler))
                else:
                    assert port is FuncPortType.const
                    if not isinstance(value, ConstExpr):
                        self.error_c(ErrorType.ARG_NOT_CONST, arg=arg)
                    self.register_symbol(arg, value)
            # Visit body
            file.write_debug("## Start of inline function")
            for stmt in func.node.body:
                self.visit(stmt)
            file.write_debug("## End of inline function")
            result = self.ctx.inline_result
            rt = func.result_type
            if result is None:
                # didn't specify result
                if rt is None or rt.matches_cls(NoneDataType):
                    result = NoneLiteral()
                else:
                    self.error_c(ErrorType.NEVER_RESULT)
            else:
                got = result.data_type
                if (rt is not None) and (not rt.matches(got)):
                    self.error_c(ErrorType.WRONG_RESULT_TYPE,
                                 expect=str(rt), got=str(got))
        return result, file.commands

    def ccall_const_func(self, func: AcaciaCTFunction,
                         arg2value: Dict[str, "CTExpr"]) -> CTObj:
        with self.set_ctx(func.context):
            ctexec = CTExecuter(self.ctx.scope, self, self.file_name)
            for arg, value in arg2value.items():
                if isinstance(value, CTObj):
                    value = CTObjPtr(value)
                ctexec.current_scope.set(arg, value)
            for stmt in func.node.body:
                ctexec.visittop(stmt)
            result = ctexec.result
        if result is None:
            result = NoneLiteral()
        if not func.result_type.is_typeof(result):
            # This function might be called by `CTExecuter` and
            # therefore the error location should be set by that.
            raise Error(
                ErrorType.WRONG_RESULT_TYPE,
                expect=func.result_type.name,
                got=abs(result).cdata_type.name
            )
        return result

    # subscript

    def visit_Subscript(self, node: Subscript):
        object_: AcaciaExpr = self.visit(node.object)
        getitem = self.attribute_of(object_, "__getitem__")
        if getitem is None:
            self.error_c(ErrorType.NO_GETITEM,
                         type_=str(object_.data_type))
        args = list(map(self.visit, node.subscripts))
        return self.call_function(
            getitem, args, {}, location=self.node_location(node)
        )



"""
Acacia - a programming language that runs in Minecraft Bedrock Edition
by compiling code into mcfunction commands, supporting variables,
functions, custom types, arithmetic, entity and coordinate system and
very rich compile-time structures like strings, lists and maps.
"""

__version__ = "0.1.0"



"""Position offset.
Builtin type that represents a position, but without the context. (i.e
the /execute subcommands)
For instance, "~ ~5 ~" is a position offset, while
"~ ~5 ~ from xxx entity" is a position.
"""

__all__ = ["PosOffsetType", "PosOffsetDataType", "PosOffset", "CoordinateType"]

from typing import List
from enum import Enum

from acaciamc.error import *
from acaciamc.tools import axe, cmethod_of, ImmutableMixin, transform_immutable
from acaciamc.mccmdgen.datatype import DefaultDataType
from acaciamc.mccmdgen.ctexpr import CTDataType
from acaciamc.mccmdgen.expr import *
from .types import Type

class CoordinateType(Enum):
    ABSOLUTE = ""
    RELATIVE = "~"
    LOCAL = "^"

class PosOffsetDataType(DefaultDataType):
    name = "Offset"

ctdt_posoffset = CTDataType("Offset")

class PosOffsetType(Type):
    def do_init(self):
        @cmethod_of(self, "__new__")
        @axe.chop
        @axe.arg("x", axe.Nullable(axe.LiteralFloat()), default=None)
        @axe.arg("y", axe.Nullable(axe.LiteralFloat()), default=None)
        @axe.arg("z", axe.Nullable(axe.LiteralFloat()), default=None)
        @axe.star
        @axe.arg("x_abs", axe.Nullable(axe.LiteralFloat()), default=None)
        @axe.arg("y_abs", axe.Nullable(axe.LiteralFloat()), default=None)
        @axe.arg("z_abs", axe.Nullable(axe.LiteralFloat()), default=None)
        def _new(compiler, x, y, z, x_abs, y_abs, z_abs):
            """
            const def __new__(
                x = None, y = None, z = None,
                *,
                x_abs = None, y_abs = None, z_abs = None
            ) -> Offset
            Return a new Offset object. All the parameters are either
            None or float. At most one of * and *_abs should be float
            for each of x, y, z. The created object will use
            relative coordinate if * is used, absolute for *_abs, or
            relative with value 0.0 if neither is given.
            Consider AbsPos if all three of them are absolute.
            """
            res = PosOffset()
            if x is not None and x_abs is not None:
                raise Error(ErrorType.POS_OFFSET_CTOR_ARG, axis="x")
            if y is not None and y_abs is not None:
                raise Error(ErrorType.POS_OFFSET_CTOR_ARG, axis="y")
            if z is not None and z_abs is not None:
                raise Error(ErrorType.POS_OFFSET_CTOR_ARG, axis="z")
            if x is None:
                x = 0.0
            if y is None:
                y = 0.0
            if z is None:
                z = 0.0
            res._set(CoordinateType.RELATIVE, x, y, z)
            res._set(CoordinateType.ABSOLUTE, x_abs, y_abs, z_abs)
            return res
        @cmethod_of(self, "local")
        @axe.chop
        @axe.arg("left", axe.LiteralFloat(), default=0.0)
        @axe.arg("up", axe.LiteralFloat(), default=0.0)
        @axe.arg("front", axe.LiteralFloat(), default=0.0)
        def _local(compiler, left: float, up: float, front: float):
            """Offset.local(left, up, front)
            Return new object using local coordinate.
            """
            return PosOffset.local(left, up, front)

    def datatype_hook(self):
        return PosOffsetDataType()

    def cdatatype_hook(self):
        return ctdt_posoffset

class PosOffset(ConstExprCombined, ImmutableMixin):
    cdata_type = ctdt_posoffset

    def __init__(self):
        super().__init__(PosOffsetDataType())
        self.values: List[float] = [0.0, 0.0, 0.0]
        self.value_types: List[CoordinateType] = \
            [CoordinateType.RELATIVE for _ in range(3)]
        @cmethod_of(self, "abs")
        @axe.chop
        @axe.arg("x", axe.Nullable(axe.PosXZ()), default=None)
        @axe.arg("y", axe.Nullable(axe.LiteralFloat()), default=None)
        @axe.arg("z", axe.Nullable(axe.PosXZ()), default=None)
        @transform_immutable(self)
        def _abs(self: PosOffset, compiler, x, y, z):
            self._set(CoordinateType.ABSOLUTE, x, y, z)
            return self
        @cmethod_of(self, "offset")
        @axe.chop
        @axe.arg("x", axe.Nullable(axe.LiteralFloat()), default=None)
        @axe.arg("y", axe.Nullable(axe.LiteralFloat()), default=None)
        @axe.arg("z", axe.Nullable(axe.LiteralFloat()), default=None)
        @transform_immutable(self)
        def _offset(self: PosOffset, compiler, x, y, z):
            self._set(CoordinateType.RELATIVE, x, y, z)
            return self
        """
        .offset(x, y, z) .abs(x, y, z)
        "x", "y" and "z" are either "None" or int literal or float.
        "offset" sets given axes to use relative coordinate,
        while "abs" sets them to use absolute coordinate.
        "abs" rounds integer x and z value to block center.
        """

    def __str__(self) -> str:
        return " ".join(
            type_.value + str(value)
            for type_, value in zip(self.value_types, self.values)
        )

    def _set(self, coord_type: CoordinateType, x, y, z):
        if x is not None:
            self.set(0, x, coord_type)
        if y is not None:
            self.set(1, y, coord_type)
        if z is not None:
            self.set(2, z, coord_type)

    @classmethod
    def local(cls, left: float, up: float, front: float):
        inst = cls()
        inst.set(0, left, CoordinateType.LOCAL)
        inst.set(1, up, CoordinateType.LOCAL)
        inst.set(2, front, CoordinateType.LOCAL)
        return inst

    def set(self, index: int, value: float, type_: CoordinateType):
        assert 0 <= index <= 2
        self.values[index] = value
        self.value_types[index] = type_

    def copy(self) -> "PosOffset":
        res = PosOffset()
        res.values = self.values.copy()
        res.value_types = self.value_types.copy()
        return res



"""Low-level utility for Minecraft selectors."""

__all__ = ["MCSelector", "SELECTORVAR_T"]

from typing import Union, Dict, Any, Optional, List
from copy import deepcopy

from acaciamc.mccmdgen.cmds import mc_str

SELECTORVAR_T = str  # Literal["a", "e", "r", "p", "s", "initiator"]

class MCSelector:
    """Low-level utility for Minecraft selectors."""
    def __init__(self, var: Union[SELECTORVAR_T, None] = None) -> None:
        # When `var` is None, the variable is unknown.
        self.var = var
        self.args: Dict[str, Any] = {}

    def copy(self) -> "MCSelector":
        res = MCSelector(self.var)
        res.args = deepcopy(self.args)
        return res

    def is_var_set(self) -> bool:
        return self.var is not None

    def has_arg(self, arg: str) -> bool:
        return arg in self.args

    def get_tags(self) -> List[str]:
        """
        Return the list of tags in the selector.
        Modifying the returned list will modify the selector.
        """
        return self.args.setdefault("tag", [])

    @staticmethod
    def arg_to_str(arg: str, value):
        if arg in ("type", "name", "m"):
            return "%s=%s" % (arg, value)
        elif arg in ("tag", "family"):
            return ",".join(("%s=%s" % (arg, v)) for v in value)
        elif arg in ("tag!", "type!", "name!", "family!", "m!"):
            n = arg[:-1]  # strip "!"
            return ",".join(("%s=!%s" % (n, v)) for v in value)
        elif arg in ("c", "l", "lm"):
            return "%s=%d" % (arg, value)
        elif arg in ("dx", "dy", "dz", "rx", "rxm", "ry", "rym", "r", "rm"):
            return "%s=%.3f" % (arg, value)
        elif arg == "hasitem":
            res = []
            for defi in value:
                options = [
                    "item=%s" % defi["item"],
                    "quantity=%s" % defi["quantity"]
                ]
                if "slot_type" in defi:
                    options.append("location=%s" % defi["slot_type"])
                    if "slot_num" in defi:
                        options.append("slot=%s" % defi["slot_num"])
                if "data" in defi:
                    options.append("data=%d" % defi["data"])
                res.append("{%s}" % ",".join(options))
            if len(res) == 1:
                return "hasitem=%s" % res[0]
            else:
                return "hasitem=[%s]" % ",".join(res)
        elif arg == "scores":
            return "scores={%s}" % ",".join(
                "%s=%s" % (scb, range_) for scb, range_ in value
            )
        elif arg == "haspermission":
            return "haspermission={%s}" % ",".join(
                [("%s=enabled" % p) for p in value[0]]
                + [("%s=disabled" % p) for p in value[1]]
            )
        else:
            raise ValueError("Unknown selector argument: %r" % arg)

    def to_str(self):
        var = self.var
        if var is None:
            var = "e"  # when var is not set it should be all entities
        return "@%s%s" % (
            var,
            "[%s]" % ",".join(self.arg_to_str(arg, value)
                              for arg, value in self.args.items())
            if self.args else ""
        )

    def player_type(self):
        if self.has_arg("type"):
            if self.args["type"] not in ("player", "minecraft:player"):
                raise ValueError
        else:
            self.type("player")

    def tag(self, *tag: str):
        if not self.has_arg("tag"):
            self.args["tag"] = []
        self.args["tag"].extend(map(mc_str, tag))

    def tag_n(self, *tag: str):
        if not self.has_arg("tag!"):
            self.args["tag!"] = []
        self.args["tag!"].extend(map(mc_str, tag))

    def type(self, type_: str):
        self.args["type"] = type_

    def type_n(self, *types: str):
        if not self.has_arg("type!"):
            self.args["type!"] = []
        self.args["type!"].extend(types)

    def family(self, *families: str):
        if not self.has_arg("family"):
            self.args["family"] = []
        self.args["family"].extend(families)

    def family_n(self, *families: str):
        if not self.has_arg("family!"):
            self.args["family!"] = []
        self.args["family!"].extend(families)

    def limit(self, limit: int):
        self.args["c"] = limit

    def distance(self, min_: Optional[float], max_: Optional[float]):
        if min_ is not None:
            self.args["rm"] = min_
        if max_ is not None:
            self.args["r"] = max_

    def volume(self, dx: float, dy: float, dz: float):
        self.args["dx"] = dx
        self.args["dy"] = dy
        self.args["dz"] = dz

    def rot_vertical(self, min_: float, max_: float):
        self.args["rxm"] = min_
        self.args["rx"] = max_

    def rot_horizontal(self, min_: float, max_: float):
        self.args["rym"] = min_
        self.args["ry"] = max_

    def name(self, name: str):
        self.args["name"] = mc_str(name)

    def name_n(self, *names: str):
        if not self.has_arg("name!"):
            self.args["name!"] = []
        self.args["name!"].extend(map(mc_str, names))

    def has_item(self, item: str, quantity: str, data: Optional[int],
                 slot_type: Optional[str], slot_num: Optional[int]):
        if not self.has_arg("hasitem"):
            self.args["hasitem"] = []
        v = {"item": item, "quantity": quantity}
        if data is not None:
            v["data"] = data
        if slot_type is not None:
            v["slot_type"] = slot_type
            if slot_num is not None:
                v["slot_num"] = slot_num
        self.args["hasitem"].append(v)

    def scores(self, objective: str, range_: str):
        if not self.has_arg("scores"):
            self.args["scores"] = []
        self.args["scores"].append((mc_str(objective), range_))

    def level(self, min_: Optional[int], max_: Optional[int]):
        if min_ is not None:
            self.args["lm"] = min_
        if max_ is not None:
            self.args["l"] = max_

    def game_mode(self, mode: str):
        self.args["m"] = mode

    def game_mode_n(self, *modes: str):
        if not self.has_arg("m!"):
            self.args["m!"] = []
        self.args["m!"].extend(modes)

    def has_permission(self, *permissions: str):
        if not self.has_arg("haspermission"):
            self.args["haspermission"] = ([], [])
        self.args["haspermission"][0].extend(permissions)

    def has_permission_n(self, *permissions: str):
        if not self.has_arg("haspermission"):
            self.args["haspermission"] = ([], [])
        self.args["haspermission"][1].extend(permissions)



__all__ = ['CTExecuter']

from typing import TYPE_CHECKING, Optional, List, Dict
from contextlib import contextmanager

from acaciamc.ast import *
from acaciamc.error import Error, ErrorType, SourceLocation
from acaciamc.mccmdgen.ctexpr import *
from acaciamc.mccmdgen.symbol import SymbolTable, CTRTConversionError
from acaciamc.mccmdgen.utils import unreachable, InvalidOpError
from acaciamc.objects import (
    IntLiteral, BoolLiteral, String, CTList, CTMap, NoneLiteral, Float
)

if TYPE_CHECKING:
    from acaciamc.mccmdgen.generator import Generator

BINOPS = {
    Operator.add: 'cadd',
    Operator.minus: 'csub',
    Operator.multiply: 'cmul',
    Operator.divide: 'cdiv',
    Operator.mod: 'cmod'
}
RBINOPS = {
    Operator.add: 'cradd',
    Operator.minus: 'crsub',
    Operator.multiply: 'crmul',
    Operator.divide: 'crdiv',
    Operator.mod: 'crmod'
}

class CTExecuter(ASTVisitor):
    def __init__(self, scope: SymbolTable, generator: "Generator",
                 file_name: str):
        super().__init__()
        self.generator = generator
        self.compiler = generator.compiler
        self.file_name = file_name
        self.builtins = generator.compiler.builtins
        self.current_scope = SymbolTable(scope, self.builtins)
        self.result: Optional[CTExpr] = None
        self.curnode: Optional[AST] = None

    def visit(self, node: AST, **kwds):
        oldnode = self.curnode
        self.curnode = node
        res = super().visit(node, **kwds)
        self.curnode = oldnode
        return res

    def visittop(self, node: AST, **kwds):
        try:
            return self.visit(node, **kwds)
        except Error as e:
            if not e.location.linecol_set():
                e.location.linecol = (self.curnode.lineno, self.curnode.col)
            raise

    def node_location(self, node: AST):
        return SourceLocation(self.file_name, (node.lineno, node.col))

    def error_node(self, node: AST, *args, **kwds):
        err = Error(*args, **kwds)
        # `err.location.file` is to be done by `Generator`
        err.location.linecol = (node.lineno, node.col)
        raise err

    def general_visit(self, node: AST):
        self.error_node(node, ErrorType.INVALID_CONST_STMT)

    @contextmanager
    def new_scope(self):
        self.current_scope = SymbolTable(self.current_scope, self.builtins)
        yield
        self.current_scope = self.current_scope.outer

    def register_symbol(self, name: str, value: CTExpr):
        v = self.current_scope.clookup(
            name, use_outer=False, use_builtins=False
        )
        if v is not None:
            self.error_node(self.curnode, ErrorType.SHADOWED_NAME, name=name)
        self.current_scope.set(name, value)

    ## Expression visitors

    def visit_Literal(self, node: Literal):
        value = node.value
        # NOTE Python bool is a subclass of int!!!
        if isinstance(value, bool):
            return BoolLiteral(value)
        elif isinstance(value, int):
            return IntLiteral(value)
        elif value is None:
            r = NoneLiteral()
            r.is_temporary = True
            return r
        elif isinstance(value, float):
            return Float(value)
        unreachable()

    def visit_StrLiteral(self, node: StrLiteral):
        s = self.visit(node.content)
        return String(s)

    def visit_Self(self, node: Self):
        self.error_node(node, ErrorType.SELF_OUT_OF_SCOPE)

    def visit_ListDef(self, node: ListDef):
        return CTList(map(self.visit, node.items))

    def visit_MapDef(self, node: MapDef):
        return CTMap(map(self.visit, node.keys),
                     map(self.visit, node.values))

    def visit_Identifier(self, node: Identifier):
        name = node.name
        try:
            v = self.current_scope.clookup(name)
        except CTRTConversionError as err:
            self.error_node(node, ErrorType.NOT_CONST_NAME, name=name,
                            type_=str(err.expr.data_type))
        if v is None:
            self.error_node(node, ErrorType.NAME_NOT_DEFINED, name=name)
        return v

    def attribute_of(self, primary: CTExpr, attr: str):
        primary = abs(primary)
        try:
            v = primary.attributes.clookup(attr)
        except CTRTConversionError as err:
            self.error_node(self.curnode, ErrorType.NOT_CONST_ATTR,
                            attr=attr, type_=str(err.expr.data_type),
                            primary=primary.cdata_type.name)
        if v is None:
            self.error_node(
                self.curnode,
                ErrorType.HAS_NO_ATTRIBUTE,
                value_type=primary.cdata_type.name, attr=attr
            )
        return v

    def visit_Attribute(self, node: Attribute):
        obj: CTObj = abs(self.visit(node.object))
        return self.attribute_of(obj, node.attr)

    def visit_UnaryOp(self, node: UnaryOp):
        obj: CTObj = abs(self.visit(node.operand))
        if node.operator is Operator.not_:
            meth = obj.cunarynot
        elif node.operator is Operator.positive:
            meth = obj.cunarypos
        elif node.operator is Operator.negative:
            meth = obj.cunaryneg
        else:
            unreachable()
        try:
            return meth()
        except InvalidOpError:
            self.error_node(node, ErrorType.INVALID_OPERAND,
                            operator=node.operator.value,
                            operand=f'"{obj.cdata_type.name}"')

    def visit_BinOp(self, node: BinOp):
        left: CTObj = abs(self.visit(node.left))
        right: CTObj = abs(self.visit(node.right))
        try:
            res = getattr(left, BINOPS[node.operator])(right)
        except InvalidOpError:
            try:
                res = getattr(right, RBINOPS[node.operator])(left)
            except InvalidOpError:
                ls = left.cdata_type.name
                rs = right.cdata_type.name
                self.error_node(
                    node, ErrorType.INVALID_OPERAND,
                    operator=node.operator.value,
                    operand=f'"{ls}", "{rs}"'
                )
        return res

    def visit_CompareOp(self, node: CompareOp):
        operands = map(self.visit, node.operands)
        left, right = None, abs(self.visit(node.left))
        final = True
        for operator, operand in zip(node.operators, operands):
            left, right = right, abs(operand)
            try:
                res = left.ccompare(operator, right)
            except InvalidOpError:
                try:
                    res = right.ccompare(COMPOP_SWAP[operator], left)
                except InvalidOpError:
                    ls = left.cdata_type.name
                    rs = right.cdata_type.name
                    self.error_node(
                        node, ErrorType.INVALID_OPERAND,
                        operator=operator.value,
                        operand=f'"{ls}", "{rs}"'
                    )
            if not res:
                final = False
        return BoolLiteral(final)

    def visit_BoolOp(self, node: BoolOp):
        operands: List[CTObj] = [
            abs(self.visit(operand)) for operand in node.operands
        ]
        operator = node.operator
        res = operator is Operator.and_
        for i, operand in enumerate(operands):
            if not isinstance(operand, BoolLiteral):
                self.error_node(
                    node.operands[i],
                    ErrorType.INVALID_BOOLOP_OPERAND,
                    operator=operator.value,
                    operand=operand.cdata_type.name
                )
            if operand.value and operator is Operator.or_:
                res = True
            elif not operand.value and operator is Operator.and_:
                res = False
        return BoolLiteral(res)

    def visit_Call(self, node: Call):
        func: CTObj = abs(self.visit(node.func))
        if not isinstance(func, CTCallable):
            self.error_node(node, ErrorType.UNCALLABLE,
                            expr_type=func.cdata_type.name)
        args, keywords = self.visit(node.table)
        return func.ccall_withframe(
            args, keywords, self.compiler,
            self.node_location(node)
        )

    def visit_Subscript(self, node: Subscript):
        obj: CTObj = abs(self.visit(node.object))
        subscripts = list(map(self.visit, node.subscripts))
        meth = self.attribute_of(obj, "__ct_getitem__")
        if meth is None:
            self.error_node(node, ErrorType.NO_GETITEM,
                            type_=obj.cdata_type.name)
        meth = abs(meth)
        if not isinstance(meth, CTCallable):
            self.error_node(node, ErrorType.UNCALLABLE,
                            expr_type=obj.cdata_type.name)
        return meth.ccall_withframe(
            subscripts, {}, self.compiler,
            self.node_location(node)
        )

    ## Statement visitors

    def visit_VarDef(self, node: VarDef):
        if node.value is None:
            self.error_node(node, ErrorType.UNINITIALIZED_CONST)
        dt: CTDataType = self.visit(node.type)
        value: CTObj = abs(self.visit(node.value))
        if not dt.is_typeof(value):
            self.error_node(node.type, ErrorType.WRONG_ASSIGN_TYPE,
                            got=value.cdata_type.name, expect=dt.name)
        self.register_symbol(node.target, CTObjPtr(value))

    def visit_AutoVarDef(self, node: AutoVarDef):
        value: CTObj = abs(self.visit(node.value))
        self.register_symbol(node.target, CTObjPtr(value))

    def visit_Assign(self, node: Assign):
        target: CTExpr = self.visit(node.target)
        if not isinstance(target, CTObjPtr):
            self.error_node(node.target, ErrorType.INVALID_ASSIGN_TARGET)
        value: CTObj = abs(self.visit(node.value))
        target.set(value)

    def visit_AugmentedAssign(self, node: AugmentedAssign):
        # A += B during compile time is implemented just as A = A + B
        target: CTExpr = self.visit(node.target)
        if not isinstance(target, CTObjPtr):
            self.error_node(node.target, ErrorType.INVALID_ASSIGN_TARGET)
        value: CTObj = abs(self.visit(node.value))
        try:
            res = getattr(abs(target), BINOPS[node.operator])(value)
        except InvalidOpError:
            t1 = abs(target).cdata_type.name
            t2 = value.cdata_type.name
            self.error_node(node, ErrorType.INVALID_OPERAND,
                            operator=f'{node.operator.value}=',
                            operand=f'"{t1}", "{t2}"')
        target.set(res)

    def visit_ReferenceDef(self, node: ReferenceDef):
        value: CTExpr = self.visit(node.value)
        if not isinstance(value, CTObjPtr):
            self.error_node(node.value, ErrorType.CANT_REF)
        if node.type is not None:
            dt: CTDataType = self.visit(node.type)
            if not dt.is_typeof(value):
                self.error_node(node.type, ErrorType.WRONG_REF_TYPE,
                                anno=dt.name, got=value.cdata_type.name)
        self.register_symbol(node.name, value)

    def visit_ExprStatement(self, node: ExprStatement):
        self.visit(node.value)

    def visit_Pass(self, node: Pass):
        pass

    def visit_If(self, node: If):
        condition: CTObj = abs(self.visit(node.condition))
        if not isinstance(condition, BoolLiteral):
            self.error_node(node.condition, ErrorType.WRONG_IF_CONDITION,
                            got=condition.cdata_type.name)
        running = node.body if condition.value else node.else_body
        for stmt in running:
            self.visit(stmt)

    def visit_While(self, node: While):
        def _while_cond() -> bool:
            condition: CTObj = abs(self.visit(node.condition))
            if not isinstance(condition, BoolLiteral):
                self.error_node(
                    node.condition, ErrorType.WRONG_WHILE_CONDITION,
                    got=condition.cdata_type.name
                )
            return condition.value
        while _while_cond():
            for stmt in node.body:
                self.visit(stmt)

    def visit_For(self, node: For):
        iterable: CTObj = abs(self.visit(node.expr))
        try:
            values = iterable.citerate()
        except InvalidOpError:
            self.error_node(
                node.expr, ErrorType.NOT_ITERABLE,
                type_=iterable.cdata_type.name
            )
        for value in values:
            with self.new_scope():
                self.register_symbol(node.name, abs(value))
                for stmt in node.body:
                    self.visit(stmt)

    def visit_Result(self, node: Result):
        # Toplevel result
        self.result = self.visit(node.value)

    # TODO compile time struct?

    # Imports are not supported (so far) because compile time functions
    # are supposed to have no side effects, but parsing a module may
    # produce commands or change runtime variables.

    # Nested functions and entity templates are not supported (so far)
    # because using symbols from current scope in nested functions could
    # cause unexpected results (e.g. using symbols in runtime function
    # and inline/const function may lead to different results).

    ## Other visitors

    def visit_TypeSpec(self, node: TypeSpec) -> CTDataType:
        type_: CTObj = abs(self.visit(node.content))
        try:
            dt = type_.cdatatype_hook()
        except InvalidOpError:
            self.error_node(node, ErrorType.INVALID_TYPE_SPEC,
                            got=type_.cdata_type.name)
        return dt

    def visit_CallTable(self, node: CallTable):
        args: List[CTExpr] = []
        keywords: Dict[str, CTExpr] = {}
        for value in node.args:
            args.append(self.visit(value))
        for arg, value in node.keywords.items():
            keywords[arg] = self.visit(value)
        return args, keywords

    def visit_FormattedStr(self, node: FormattedStr) -> str:
        res: List[str] = []
        for section in node.content:
            if isinstance(section, str):
                res.append(section)
            else:
                expr: CTObj = abs(self.visit(section))
                try:
                    value = expr.cstringify()
                except InvalidOpError:
                    self.error_node(section, ErrorType.INVALID_FEXPR)
                res.append(value)
        return ''.join(res)



"""
The main compiler of Acacia.
It assembles files and classes together and writes output.
"""

__all__ = ['Compiler', 'Config']

from typing import (
    Tuple, Union, Optional, Callable, Dict, NamedTuple, List, TYPE_CHECKING
)
import os
from contextlib import contextmanager

from acaciamc.ast import ModuleMeta
from acaciamc.error import *
from acaciamc.tokenizer import Tokenizer
from acaciamc.parser import Parser
from acaciamc.mccmdgen.generator import Generator
from acaciamc.mccmdgen.expr import *
from acaciamc.mccmdgen.optimizer import Optimizer
from acaciamc.mccmdgen.utils import unreachable
from acaciamc.objects import (
    IntVar, BinaryModule, EntityTemplate, DEFAULT_ENTITY_NEW
)
import acaciamc.mccmdgen.cmds as cmds

if TYPE_CHECKING:
    from acaciamc.tools.versionlib import VERSION_T

class OutputManager(cmds.FunctionsManager):
    def __init__(self, cfg: "Config"):
        super().__init__(cfg.scoreboard)
        self._cfg = cfg
        self._lib_count = 0
        sep = '' if not cfg.root_folder else "/"
        self._mcfp_template = f"{cfg.root_folder}{sep}%s"
        self.tick_file_path = f"{self._cfg.internal_folder}/tick"
        self.tick_file_full_path = self.mcfunction_path(self.tick_file_path)

    def mcfunction_path(self, path: str) -> str:
        return self._mcfp_template % path

    def new_file(self, file: cmds.MCFunctionFile, path: str):
        file.set_path(self.mcfunction_path(path))
        self.add_file(file)

    def add_lib(self, file: cmds.MCFunctionFile):
        self._lib_count += 1
        fname = f"{self._cfg.internal_folder}/acalib{self._lib_count}"
        self.new_file(file, fname)

class OutputOptimized(OutputManager, Optimizer):
    def entry_files(self):
        internal = self.mcfunction_path(self._cfg.internal_folder) + "/"
        for file in self.files:
            path = file.get_path()
            if (
                not path.startswith(internal)
                or path == self.tick_file_full_path
            ):
                yield file

    @property
    def max_inline_file_size(self) -> int:
        return self._cfg.max_inline_file_size

    def dont_inline_execute_call(self, file: cmds.MCFunctionFile) -> bool:
        # Expanding /execute function calls in tick.mcfunction can
        # decrease performance badly.
        return file.get_path() == self.tick_file_full_path

class Config(NamedTuple):
    # Generate debug comments in .mcfunction files
    debug_comments: bool = False
    # Subpath of "functions" in a behavior pack where Acacia would put
    # all mcfunctions into. Must be a sequence of Acacia identifiers
    # joined with "/", or an empty string. Redundant slashes are NOT
    # allowed.
    root_folder: str = ''
    # Name of the folder that contains internal mcfunctions
    internal_folder: str = '_acacia'
    # Name of mcfunction file that executes the program
    main_file: str = 'main'
    # Prefix of scoreboard that Acacia uses to hold data
    scoreboard: str = 'acacia'
    # Prefix of entity tags
    entity_tag: str = 'acacia'
    # Minecraft version
    mc_version: "VERSION_T" = (1, 20, 20)
    # Minecraft Education Edition features
    education_edition: bool = False
    # Split initialization commands from main file
    split_init: bool = False
    # Name of init file (ignored if split_init is False)
    init_file: str = 'init'
    # Enable optimizer
    optimizer: bool = True
    # Maximum lines of commands an mcfunction can have to be inlined
    # even if it is called with /execute condition (ignored if optimizer
    # is False)
    max_inline_file_size: int = 20
    # Encoding of input and output files
    encoding: Optional[str] = None

class CachedModule(NamedTuple):
    # Path to module source file
    path: str
    # Module object
    module: AcaciaExpr
    # Module main function (None if no command is needed)
    main: Optional[cmds.MCFunctionFile] = None
    # Runtime variable to indicate if main function has been executed
    # (None if `main` is None)
    loaded_var: Optional[cmds.ScbSlot] = None

class Compiler:
    """Start compiling the project
    A Compiler manage the resources in the compile task and
    connect the steps to compile: Tokenizer -> Parser -> Generator.
    """
    def __init__(self, main_path: str, cfg: Optional[Config] = None):
        """
        main_path: path of main source file
        cfg: optional `Config` object
        """
        self.main_dir, _ = os.path.split(main_path)
        self.main_dir = os.path.realpath(self.main_dir)
        ACACIA = os.path.realpath(os.path.dirname(__file__))
        self.path = [
            os.path.join(ACACIA, 'modules'),  # buitlin modules
            self.main_dir  # find in program entry (main file)
        ]  # modules will be found in these directories
        self.cfg = Config() if cfg is None else cfg
        if self.cfg.optimizer:
            self.output_mgr = OutputOptimized(self.cfg)
        else:
            self.output_mgr = OutputManager(self.cfg)
        self.file_main = cmds.MCFunctionFile()  # load program
        self.file_tick = cmds.MCFunctionFile()  # runs every tick
        self.output_mgr.new_file(self.file_main, self.cfg.main_file)
        self.current_generator: Optional[Generator] = None
        self._interface_paths: Dict[str, SourceLocation] = {}
        self._score_max = 0  # max id of score allocated
        self._scoreboard_max = 0  # max id of scoreboard allocated
        self._entity_tag_max = 0  # max id of entity tag allocated
        self._free_tmp_score = []  # free tmp scores (see `allocate_tmp`)
        self._current_file = None  # str; Path of current parsing file
        self._cached_modules: List[CachedModule] = []
        self._loading_files = []  # paths of Acacia modules that are loading
        self._before_finish_cbs = []  # callbacks to run before finish
        self._entity_template_id_max = 0  # max id of entity template
        self.etemplate_id_scb = self.add_scoreboard()

        # --- BUILTINS ---
        self.base_template = EntityTemplate(
            name="Entity",
            field_types={}, field_metas={}, methods={},
            method_new=DEFAULT_ENTITY_NEW,
            method_qualifiers={}, parents=[], compiler=self
        )
        self.external_template = EntityTemplate(
            "ExternalEntity",
            field_types={}, field_metas={}, methods={},
            method_new=None,
            method_qualifiers={}, parents=[], compiler=self
        )
        builtin_mod = self.get_module(ModuleMeta("builtins"), self.file_main)
        self.builtins = builtin_mod.attribute_table

        # --- START COMPILE ---
        ## start
        with self._load_generator(main_path, self.file_main) as generator:
            generator.parse()
        ## callback
        for cb in self._before_finish_cbs:
            cb(self)
        ## add tick.mcfunction
        if self.file_tick.has_content():
            self.output_mgr.new_file(
                self.file_tick, self.output_mgr.tick_file_path
            )
        ## import caching system init
        mod_loaded_vars = [
            mod.loaded_var for mod in self._cached_modules
            if mod.main is not None
        ]
        if mod_loaded_vars:
            self.file_main.commands[:0] = [
                cmds.ScbSetConst(v, 0) for v in mod_loaded_vars
            ]
            self.file_main.commands.insert(
                0, cmds.Comment("# Reset import caching flags")
            )
        ## init
        init = self.output_mgr.generate_init()
        if self.cfg.split_init:
            init_file = cmds.MCFunctionFile()
            self.output_mgr.new_file(init_file, self.cfg.init_file)
            init_file.write_debug(
                '## Usage: Initialize Acacia, only need to be ran ONCE',
                '## Execute this before running anything from Acacia!!!'
            )
            init_file.extend(init)
        else:
            self.file_main.commands[:0] = init
        ## comment on main.mcfunction
        self.file_main.commands[:0] = [
            cmds.Comment('## Usage: Run this Acacia project'),
            cmds.Comment('## Execute this before using interfaces!!!')
        ]
        ## optimize
        if isinstance(self.output_mgr, OutputOptimized):
            self.output_mgr.optimize()

    def output(self, path: str):
        """
        Output result to `path`.
        e.g. when `path` is "a/b", main file is generated at
        "a/b/{self.cfg.root_folder}/main.mcfunction".
        """
        # Mcfunctions
        for file in self.output_mgr.files:
            self._write_mcfunction(file, path)
        # tick.json
        if self.file_tick.has_content():
            self._write_file(
                '{"values": ["%s"]}' % self.output_mgr.tick_file_full_path,
                os.path.join(path, 'tick.json')
            )

    def raise_error(self, error: Error):
        if self.current_generator is not None:
            self.current_generator.fix_error_location(error)
        error.location.file = self._current_file
        raise error

    def add_file(self, file: cmds.MCFunctionFile, path: Optional[str] = None):
        """
        Add a new file to the project.
        `path` is the path relative to the self.cfg.root_folder.
        If `path` is None, the file will be added as an internal lib.
        """
        if path is None:
            self.output_mgr.add_lib(file)
        else:
            self.output_mgr.new_file(file, path)

    # -- About allocation --

    def allocate(self) -> cmds.ScbSlot:
        """Apply for a new score."""
        return self.output_mgr.allocate()

    def allocate_tmp(self) -> cmds.ScbSlot:
        """Apply for a temporary score
        NOTE Only do this when you are really using a TEMPORARY var
        because the var returned might have been used by others
        NOTE Temporary vars are only available within 1 statement
        when it comes to next statement, current vars are deleted.
        """
        if self._free_tmp_score:
            # if there are free vars in list, reuse them
            res = self._free_tmp_score.pop()
        else:
            # else, allocate a new one
            res = self.allocate()
        self.current_generator.current_tmp_scores.append(res)
        return res

    def free_tmp(self, slot: cmds.ScbSlot):
        """Free the tmp var allocated by method `allocate_tmp`.
        NOTE This is called automatically.
        """
        self._free_tmp_score.append(slot)

    def add_int_const(self, value: int) -> IntVar:
        """Sometimes a constant is needed when calculating in MC
        e.g. `a * 2`, we need a score to store 2, so that we can use
             `scoreboard operation ... *= const2 ...`
        This method can create one.
        """
        return IntVar(slot=self.output_mgr.int_const(value))

    def add_scoreboard(self) -> str:
        """Apply for a new scoreboard"""
        return self.output_mgr.add_scoreboard()

    def allocate_entity_tag(self) -> str:
        """Return a new entity tag."""
        self._entity_tag_max += 1
        return self.cfg.entity_tag + str(self._entity_tag_max)

    def allocate_etemplate_id(self) -> int:
        """Return a new entity template id."""
        self._entity_template_id_max += 1
        return self._entity_template_id_max

    # -- End allocation --

    def before_finish(self, callback: Callable[["Compiler"], None]):
        """Add a callback before compilation finishes."""
        self._before_finish_cbs.append(callback)

    def find_module(self, meta: ModuleMeta) -> Union[str, None]:
        """Find a module.
        Return path of module or None is not found
        In details, this work like this:
        1. if leading_dots is 0, find the module in self.path
           else, find the module in the parent folder of
           self.main_dir; count of dots decides which parent folder
           e.g. main_dir at a/b/c; "...pack.file" -> a/pack
        2. when the start directory is decided, follow the names
           in `parents` and go deeper in folders
           e.g. main_dir at a/b/c; ".pack.sub.file" -> a/b/c/pack/sub
        3. find the module file in the directory
           e.g. main_dir at a/b/c; ".pack.file" -> a/b/c/pack/file.aca
        """
        ## Step 1
        if meta.leading_dots == 0:
            paths = self.path
        else:
            _path = self.main_dir
            for _ in range(meta.leading_dots - 1):
                _path = os.path.join(_path, os.pardir)
            paths = (_path,)
        ## Step 2~3
        for root in paths:
            ## Step 2
            final = os.path.join(root, *meta.parents)
            if not os.path.isdir(final):
                # failed to find any of the child directory,
                # meaning this `root` is invalid; continue
                continue
            ## Step 3
            for child in os.listdir(final):
                path = os.path.join(final, child)
                if not os.path.isfile(path):
                    continue
                got_name, ext = os.path.splitext(child)
                if (got_name == meta.last_name
                    and (ext == '.py' or ext == '.aca')):
                    return path
        return None

    def parse_module(self, meta: ModuleMeta,
                     mcfunc: cmds.MCFunctionFile) -> Tuple[AcaciaExpr, str]:
        """Parse and get a module and its path."""
        path = self.find_module(meta)
        if path is None:
            self.raise_error(
                Error(ErrorType.MODULE_NOT_FOUND, module=str(meta))
            )
        # Get the module accoding to path
        for cm in self._cached_modules:
            # Return cached if exists
            if os.path.samefile(cm.path, path):
                finalcm = cm
                break
        else:
            # Load the module
            _, ext = os.path.splitext(path)
            mod_main = cmds.MCFunctionFile()
            if ext == ".aca":
                # Parse the Acacia module
                with self._load_generator(path, mod_main) as generator:
                    mod = generator.parse_as_module()
            elif ext == ".py":
                # Parse the binary module
                mod = BinaryModule(path)
                mod_main.extend(mod.execute(self))
            else:
                unreachable()
            if mod_main.has_content():
                loaded_var = self.allocate()
                self.add_file(mod_main)
                mod_main.commands[:0] = [
                    cmds.Comment(f"## Main file of module {path}"),
                    cmds.ScbSetConst(loaded_var, 1)
                ]
                finalcm = CachedModule(path, mod, mod_main, loaded_var)
            else:
                finalcm = CachedModule(path, mod)
            self._cached_modules.append(finalcm)
        if finalcm.main is not None:
            mcfunc.write(cmds.Execute(
                [cmds.ExecuteScoreMatch(finalcm.loaded_var, "0")],
                cmds.InvokeFunction(finalcm.main)
            ))
        return finalcm.module, path

    def get_module(self, meta: ModuleMeta,
                   mcfunc: cmds.MCFunctionFile) -> AcaciaExpr:
        """Parse a module meta and just return the module
        An API for binary module developing.
        """
        return self.parse_module(meta, mcfunc)[0]

    def lookup_interface(self, path: str) -> Optional[SourceLocation]:
        """Return the location of the interface if it exists."""
        path = path.lower()
        if path in self._interface_paths:
            return self._interface_paths[path]
        return None

    def add_interface(self, path: str, location: SourceLocation):
        """Register an interface."""
        path = path.lower()
        self._interface_paths[path] = location

    def is_reserved_path(self, path: str) -> bool:
        """Return if a mcfunction path is reserved (not for user)."""
        checks = [self.cfg.main_file, self.cfg.internal_folder]
        if self.cfg.split_init:
            checks.append(self.cfg.init_file)
        path = path.lower()
        for check in checks:
            check = check.lower()
            if path == check or path.startswith(check + "/"):
                return True
        return False

    @contextmanager
    def _load_generator(self, path: str, mcfunc: cmds.MCFunctionFile):
        """Load the Generator of an Acacia source and store it at
        `self.current_generator`.
        """
        # Check if the module is being loading (prevent circular import)
        for p in self._loading_files:
            if os.path.samefile(p, path):
                self.raise_error(Error(ErrorType.CIRCULAR_PARSE, file_=path))
        src_file = self._open_file(path)
        oldf = self._current_file
        oldg = self.current_generator
        self._current_file = path
        self._loading_files.append(path)
        try:
            node = Parser(Tokenizer(src_file, self.cfg.mc_version)).module()
        except Error as err:
            if not err.location.file_set():
                err.location.file = path
            raise
        finally:
            src_file.close()
        self.current_generator = Generator(
            node=node, main_file=mcfunc,
            file_name=path, compiler=self
        )
        yield self.current_generator
        self._current_file = oldf
        self.current_generator = oldg
        self._loading_files.pop()

    # --- I/O Util (Internal use) ---

    def _open_file(self, path: str):
        try:
            return open(path, 'r', encoding=self.cfg.encoding)
        except Exception as err:
            self.raise_error(Error(ErrorType.IO, message=str(err)))

    def _read_file(self, path: str) -> str:
        """Read Acacia file and return source code."""
        x = self._open_file(path)
        try:
            with x:
                s = x.read()
        except Exception as err:
            self.raise_error(Error(ErrorType.IO, message=str(err)))
        else:
            return s

    def _write_file(self, content: str, path: str):
        """Write `content` to `path`."""
        os.makedirs(os.path.dirname(path), exist_ok=True)
        try:
            with open(path, 'w', encoding=self.cfg.encoding) as f:
                f.write(content)
        except Exception as err:
            self.raise_error(Error(ErrorType.IO, message=str(err)))

    def _write_mcfunction(self, file: cmds.MCFunctionFile, path: str):
        """Write content of `file` to somewhere in output `path`
        e.g. when `path` is "a/b", `file.path` is "a", file is at
        "a/b/a.mcfunction".
        """
        self._write_file(
            content=file.to_str(debugging=self.cfg.debug_comments),
            path=os.path.join(path, file.get_path() + '.mcfunction')
        )



"""Tools for working with Minecraft version related contents."""

__all__ = [
    # Utils
    "only", "format_version", "edu_only",
    # Version requirements
    "at_least", "at_most", "between", "older",
    "VersionRequirement",
    # Type checking
    "VERSION_T"
]

from typing import Callable, Tuple, Optional, TYPE_CHECKING
from abc import ABCMeta, abstractmethod

from acaciamc.error import Error as AcaciaError, ErrorType
from acaciamc.localization import localize

if TYPE_CHECKING:
    from acaciamc.compiler import Compiler

VERSION_T = Tuple[int, ...]

def format_version(version: VERSION_T):
    return ".".join(map(str, version))

class VersionRequirement(metaclass=ABCMeta):
    @abstractmethod
    def to_str(self) -> str:
        pass

    @abstractmethod
    def validate(self, version: VERSION_T) -> bool:
        pass

class VersionRanged(VersionRequirement):
    def __init__(self, min_: Optional[VERSION_T], max_: Optional[VERSION_T]):
        if not min_ and not max_:
            raise ValueError("min and max versions cannot both be None")
        self.min = min_
        self.max = max_

    def to_str(self) -> str:
        if self.min:
            if self.max:
                return "%s~%s" % (
                    format_version(self.min), format_version(self.max)
                )
            return ">=%s" % format_version(self.min)
        assert self.max
        return "<=%s" % format_version(self.max)

    def validate(self, version: VERSION_T) -> bool:
        if self.min:
            if self.max:
                return self.min <= version <= self.max
            return self.min <= version
        assert self.max
        return version <= self.max

class VersionLower(VersionRequirement):
    def __init__(self, version: VERSION_T) -> None:
        super().__init__()
        self.version = version

    def to_str(self) -> str:
        return "<%s" % format_version(self.version)

    def validate(self, version: VERSION_T) -> bool:
        return self.version < version

def at_least(version: VERSION_T) -> VersionRequirement:
    """Requires >= the given version."""
    return VersionRanged(version, None)

def at_most(version: VERSION_T) -> VersionRequirement:
    """Requires <= the given version."""
    return VersionRanged(None, version)

def between(min_: VERSION_T, max_: VERSION_T) -> VersionRequirement:
    """Requires between the given versions (inclusive)."""
    return VersionRanged(min_, max_)

def older(version: VERSION_T):
    """Requires < the given version."""
    return VersionLower(version)

def only(version: VersionRequirement):
    """Return a decorator that makes the decorated binary function only
    available when given version requirement is satisfied.
    """
    def _decorator(func: Callable):
        def _decorated(compiler: "Compiler", args, kwds):
            if version.validate(compiler.cfg.mc_version):
                return func(compiler, args, kwds)
            msg = (
                localize("tools.versionlib.only.error")
                .format(got=format_version(compiler.cfg.mc_version),
                        expected=version.to_str())
            )
            raise AcaciaError(ErrorType.ANY, message=msg)
        return _decorated
    return _decorator

def edu_only(func: Callable):
    """
    Decorator that makes a binary function only available when
    Education Edition features are enabled.
    """
    def _decorated(compiler: "Compiler", args, kwds):
        if not compiler.cfg.education_edition:
            raise AcaciaError(
                ErrorType.ANY,
                message=localize("tools.versionlib.eduonly.error")
            )
        return func(compiler, args, kwds)
    return _decorated



"""Builtin support for rotations.
NOTE When generating /execute context for rotations, we do NOT
overwrite executing position, so "facing <position>" is not
implemented. Workaround: summon an entity and teleport it to
facing position. Then use "Rot.face_entity".
"""

__all__ = ["RotType", "RotDataType", "Rotation"]

from typing import List, TYPE_CHECKING

from acaciamc.tools import axe, cmethod_of, ImmutableMixin, transform_immutable
from acaciamc.constants import DEFAULT_ANCHOR
from acaciamc.mccmdgen.datatype import DefaultDataType
from acaciamc.mccmdgen.ctexpr import CTDataType
from acaciamc.mccmdgen.expr import *
import acaciamc.mccmdgen.cmds as cmds
from . import entity as entity_module
from .types import Type
from .functions import BinaryCTFunction

if TYPE_CHECKING:
    from .entity import _EntityBase
    from acaciamc.mccmdgen.cmds import _ExecuteSubcmd

class RotDataType(DefaultDataType):
    name = "Rot"

ctdt_rotation = CTDataType("Rot")

class RotType(Type):
    def do_init(self):
        @cmethod_of(self, "__new__")
        class _new(metaclass=axe.OverloadChopped):
            """
            Rot(entity): rotation of an entity.
            Rot(int-literal, int-literal): absolute rotation
            """
            @axe.overload
            @axe.arg("entity", entity_module.EntityDataType)
            def from_entity(cls, compiler, entity: "_EntityBase"):
                inst = Rotation()
                inst.context.append(cmds.ExecuteEnv(
                    "rotated", "as " + entity.to_str()
                ))
                return inst

            @axe.overload
            @axe.arg("vertical", axe.LiteralFloat())
            @axe.arg("horizontal", axe.LiteralFloat())
            def absolute(cls, compiler, vertical: float, horizontal: float):
                inst = Rotation()
                inst.context.append(cmds.ExecuteEnv(
                    "rotated", "%s %s" % (vertical, horizontal)
                ))
                return inst
        @cmethod_of(self, "face_entity")
        @axe.chop
        @axe.arg("target", entity_module.EntityDataType)
        @axe.arg("anchor", axe.LiteralString(), default=DEFAULT_ANCHOR)
        def _face_entity(compiler, target: "_EntityBase", anchor: str):
            inst = Rotation()
            inst.context.append(cmds.ExecuteEnv(
                "facing", "entity %s %s" % (target, anchor)
            ))
            return inst

    def datatype_hook(self):
        return RotDataType()

    def cdatatype_hook(self):
        return ctdt_rotation

class Rotation(ConstExprCombined, ImmutableMixin):
    cdata_type = ctdt_rotation

    def __init__(self):
        super().__init__(RotDataType())
        self.context: List["_ExecuteSubcmd"] = []

        _abs = self._create_setter("")
        _offset = self._create_setter("~")
        """.abs(vertical, horizontal) .offset(vertical, horizontal)
        "vertical" & "horizontal" are either "None" or int literal or
        float, representing xrot and yrot values. "abs" directly sets
        rotation and "offset" rotates relatively.
        """
        self.attribute_table.set("abs", BinaryCTFunction(_abs))
        self.attribute_table.set("offset", BinaryCTFunction(_offset))

    def copy(self):
        res = Rotation()
        res.context.extend(self.context)
        return res

    def _create_setter(self, type_prefix: str):
        @axe.chop
        @axe.arg("vertical", axe.Nullable(axe.LiteralFloat()), default=None)
        @axe.arg("horizontal", axe.Nullable(axe.LiteralFloat()), default=None)
        @transform_immutable(self)
        def _setter(self: Rotation, compiler, vertical, horizontal):
            vh: List[str] = []
            for arg in (vertical, horizontal):
                if arg is None:
                    vh.append("~")
                else:
                    vh.append(type_prefix + str(arg))
            self.context.append(cmds.ExecuteEnv("rotated", " ".join(vh)))
            return self
        return _setter



"""Constants for Acacia."""

__all__ = [
    "INT_MIN", "INT_MAX", "DEFAULT_ANCHOR", "XYZ",
    "TERMINATOR_CHARS", "FUNCTION_PATH_CHARS",
    "COLORS", "COLORS_NEW"
]

from string import ascii_letters, digits

INT_MIN = -2 ** 31
INT_MAX = 2 ** 31 - 1
DEFAULT_ANCHOR = "feet"
XYZ = ("x", "y", "z")

# These characters are not allowed in strings in commands, unless they
# are quoted.
TERMINATOR_CHARS = frozenset(" ,@~^/$&\"'!#%+*=[{]}\\|<>`\n")
# These are the only characters allowed in function paths.
FUNCTION_PATH_CHARS = frozenset(ascii_letters + digits + ".(-)_/")

COLORS = {
    "black": "0",
    "dark_blue": "1",
    "dark_green": "2",
    "dark_aqua": "3",
    "dark_red": "4",
    "dark_purple": "5",
    "gold": "6",
    "gray": "7",
    "dark_gray": "8",
    "blue": "9",
    "green": "a",
    "aqua": "b",
    "red": "c",
    "light_purple": "d",
    "yellow": "e",
    "white": "f",
    "minecoin_gold": "g",
}
# 1.19.80+ colors
COLORS_NEW = {
    "material_quartz": "h",
    "material_iron": "i",
    "material_netherite": "j",
    "material_redstone": "m",
    "material_copper": "n",
    "material_gold": "p",
    "material_emerald": "q",
    "material_diamond": "s",
    "material_lapis": "t",
    "material_amethyst": "u",
}



"""Argument Axe - argument parsing tool for binary functions"""

__all__ = [
    # Decorators
    "chop", "arg", "slash", "star", "star_arg", "kwds",
    # Constants
    "POSITIONAL",
    # Overload interfaces
    "OverloadChopped", "overload", "overload_versioned",
    # Converters
    "Converter", "AnyValue", "Typed", "Multityped", "LiteralInt",
    "LiteralFloat", "LiteralString", "LiteralBool", "Nullable", "AnyOf",
    "Iterator", "Selector", "LiteralIntEnum", "LiteralStringEnum", "ListOf",
    "MapOf", "PlayerSelector", "RangedLiteralInt", "Callable", "PosXZ",
    "CTConverter", "UConverter", "CTTyped", "CTIterator", "CTReference",
    "Constant", "AnyRT",
    # Exception
    "ChopError", "ArgumentError"
]

from typing import (
    Callable as PyCallable, Optional, Type, Any, Union, List, Tuple, Dict,
    Iterable, TYPE_CHECKING
)
from itertools import chain
from functools import partial
import inspect

# `import acaciamc.objects as objects` won't work in 3.6
# because of a Python bug (see https://bugs.python.org/issue23203)
from acaciamc import objects
from acaciamc.mccmdgen import ctexpr as acaciact, expr as acacia
from acaciamc.mccmdgen.datatype import DataType
from acaciamc.mccmdgen.utils import InvalidOpError
from acaciamc.error import Error as AcaciaError, ErrorType
from acaciamc.localization import localize

if TYPE_CHECKING:
    from acaciamc.compiler import Compiler
    from acaciamc.tools.versionlib import VersionRequirement
    from acaciamc.mccmdgen.mcselector import MCSelector

### Exception

class ChopError(Exception):
    """Wrong use of Axe."""
    pass

class ArgumentError(Exception):
    """Can be raised when user is not satisfied with an argument in
    an implementation of binary function that is managed with Axe.
    """
    def __init__(self, arg: str, message: str) -> None:
        """Report error on `arg`."""
        super().__init__(arg, message)
        self.arg = arg
        self.message = message

class _WrongArgTypeError(Exception):
    pass

_PE_NOT_CONST = 1
_PE_NOT_RT = 2

class _PreconvertError(Exception):
    def __init__(self, code: int, origin: "_EXPR_T"):
        super().__init__(code, origin)
        self.code = code
        self.origin = origin

    def show(self, arg: str) -> str:
        if self.code == _PE_NOT_CONST:
            s = localize("axe.preconverterror.notconst")
        elif self.code == _PE_NOT_RT:
            s = localize("axe.preconverterror.notrt")
        else:
            raise ValueError(f'unknown error code {self.code}')
        return s.format(arg=arg, type=_exprrepr(self.origin))

### Building Stage

class _PositionalType:
    pass

POSITIONAL = _PositionalType()

_BP_ARG = 1
_BP_STAR = 2
_BP_STAR_ARG = 3
_BP_SLASH = 4
_BP_KWDS = 5

_EXPR_T = Union[acacia.AcaciaExpr, acaciact.CTExpr]
_CONVERTER_T = Union["Converter", "CTConverter"]
_RENAME_T = Union[str, _PositionalType]

class _Argument:
    def __init__(self, name: str, rename: _RENAME_T, converter: _CONVERTER_T):
        self.name = name
        self.rename = rename
        self.converter = converter

    def set_default(self, value):
        self._default = value

    def has_default(self):
        return hasattr(self, "_default")

    def get_default(self):
        if self.has_default():
            return self._default
        raise ValueError("don't have default value")

class _ArgumentList:
    def __init__(self, name: str, rename: _RENAME_T, converter: _CONVERTER_T):
        self.name = name
        self.rename = rename
        self.converter = converter

class _BuildingParser(List[Tuple[int, Any]]):
    def __init__(self, target: PyCallable):
        super().__init__()
        self._target = target

    def push_component(self, type_: int, definition=None):
        self.insert(0, (type_, definition))

    def get_target(self):
        return self._target

def _parser_component(func: PyCallable[[_BuildingParser], Any]):
    """Decorator that creates decorators that build the parser.
    If the target function returns None, the building parser will
    then be returned in decorated decorator.
    """
    def _decorated(user_input: Union[PyCallable, _BuildingParser]):
        if isinstance(user_input, _BuildingParser):
            p = user_input
        else:
            assert callable(user_input)
            p = _BuildingParser(user_input)
        res = func(p)
        if res is None:
            return p
        else:
            return res
    return _decorated

### Building Stage Interface

_TYPED_TYPE = Union[DataType, Type[DataType]]
_ARG_TYPE = Union[_TYPED_TYPE, _CONVERTER_T]
_NO_DEFAULT = object()

def _converter(type_: _ARG_TYPE):
    if isinstance(type_, (Converter, CTConverter)):
        return type_
    elif isinstance(type_, acaciact.CTDataType):
        return CTTyped(type_)
    else:
        return Typed(type_)

def arg(name: str, type_: _ARG_TYPE, rename: Optional[_RENAME_T] = None,
        default: Any = _NO_DEFAULT):
    """Return a decorator for adding an argument"""
    if rename is None:
        rename = name
    converter = _converter(type_)
    definition = _Argument(name, rename, converter)
    if default is not _NO_DEFAULT:
        definition.set_default(default)
    @_parser_component
    def _decorator(building: _BuildingParser):
        building.push_component(_BP_ARG, definition)
    return _decorator

@_parser_component
def slash(building: _BuildingParser):
    """Indicate arguments before are position-only."""
    building.push_component(_BP_SLASH)

@_parser_component
def star(building: _BuildingParser):
    """Indicate arguments after are keyword-only."""
    building.push_component(_BP_STAR)

def _arg_list(bd_type: int, name: str, type_: _ARG_TYPE,
              rename: Optional[_RENAME_T] = None):
    if rename is None:
        rename = name
    converter = _converter(type_)
    @_parser_component
    def _decorator(building: _BuildingParser):
        building.push_component(
            bd_type, _ArgumentList(name, rename, converter)
        )
    return _decorator

def star_arg(name: str, type_: _ARG_TYPE, rename: Optional[_RENAME_T] = None):
    """Catch all positional arguments. Arguments after are keyword-only."""
    return _arg_list(_BP_STAR_ARG, name, type_, rename)

def kwds(name: str, type_: _ARG_TYPE, rename: Optional[_RENAME_T] = None):
    """Catch all keyword arguments. Must at the end."""
    return _arg_list(_BP_KWDS, name, type_, rename)

### Argument converter

class Converter:
    def wrong_argument(self):
        """Used by `convert` method to raise an error."""
        raise _WrongArgTypeError

    def get_show_name(self) -> str:
        """Get show name."""
        raise NotImplementedError

    def convert(self, origin: acacia.AcaciaExpr):
        """Converts the argument.
        Can use `wrong_argument` method when failed.
        """
        raise NotImplementedError

class CTConverter:
    def wrong_argument(self):
        """Used by `convert` method to raise an error."""
        raise _WrongArgTypeError

    def crepr(self) -> str:
        raise NotImplementedError

    def cconvert(self, origin: acaciact.CTExpr):
        raise NotImplementedError

class UConverter(CTConverter, Converter):
    unbox_ctptr = True

    def crepr(self) -> str:
        return self.get_show_name()

    def uconvert(self, origin: _EXPR_T):
        raise NotImplementedError

    def convert(self, origin: acacia.AcaciaExpr):
        return self.uconvert(origin)

    def cconvert(self, origin: acaciact.CTExpr):
        if self.unbox_ctptr:
            origin = abs(origin)
        return self.uconvert(origin)

class AnyValue(UConverter):
    """Accepts any value."""
    def get_show_name(self) -> str:
        return "Any"

    def uconvert(self, origin):
        return origin

class Constant(CTConverter):
    def crepr(self) -> str:
        return "Any"

    def cconvert(self, origin):
        return origin

class AnyRT(Converter):
    def get_show_name(self) -> str:
        return "Any"

    def convert(self, origin: acacia.AcaciaExpr):
        return origin

def _type_checker(value: acacia.AcaciaExpr, type_: _TYPED_TYPE):
    if isinstance(type_, DataType):
        return type_.is_type_of(value)
    else:
        assert issubclass(type_, DataType)
        return value.data_type.matches_cls(type_)

def _type_to_str(type_: _TYPED_TYPE):
    if isinstance(type_, DataType):
        return str(type_)
    else:
        return type_.name_no_generic()

class Typed(Converter):
    """Accepts value of specified data type."""
    def __init__(self, type_: _TYPED_TYPE):
        super().__init__()
        self.type = type_

    def get_show_name(self) -> str:
        return _type_to_str(self.type)

    def convert(self, origin: acacia.AcaciaExpr):
        if not _type_checker(origin, self.type):
            self.wrong_argument()
        return origin

class Multityped(Converter):
    """Accepts value of several types."""
    def __init__(self, types: Iterable[_TYPED_TYPE]):
        super().__init__()
        self.types = tuple(types)

    def get_show_name(self) -> str:
        return " / ".join(map(_type_to_str, self.types))

    def convert(self, origin: acacia.AcaciaExpr):
        if not any(map(partial(_type_checker, origin), self.types)):
            self.wrong_argument()
        return origin

class CTTyped(CTConverter):
    """Accepts value of specified data type(s)."""
    def __init__(self, *types: acaciact.CTDataType):
        super().__init__()
        self.types = types

    def crepr(self) -> str:
        return " / ".join(dt.name for dt in self.types)

    def cconvert(self, origin: acaciact.CTExpr):
        if not any(dt.is_typeof(origin) for dt in self.types):
            self.wrong_argument()
        return origin

class UTyped(UConverter):
    """Accepts value of specified data type."""
    def __init__(self, rtype: _TYPED_TYPE, ctype: acaciact.CTDataType):
        super().__init__()
        self.rtype = rtype
        self.ctype = ctype

    def get_show_name(self) -> str:
        return _type_to_str(self.rtype)

    def uconvert(self, origin: _EXPR_T):
        if isinstance(origin, acaciact.CTObj):
            if not self.ctype.is_typeof(origin):
                self.wrong_argument()
        else:
            if not _type_checker(origin, self.rtype):
                self.wrong_argument()
        return origin

class UMultityped(UConverter):
    """Accepts value of several types."""
    def __init__(self, rtypes: Iterable[_TYPED_TYPE],
                 ctypes: Iterable[acaciact.CTDataType]):
        super().__init__()
        self.rtypes = tuple(rtypes)
        self.ctypes = tuple(ctypes)

    def get_show_name(self) -> str:
        return " / ".join(map(_type_to_str, self.rtypes))

    def uconvert(self, origin: _EXPR_T):
        if isinstance(origin, acaciact.CTObj):
            if not any(dt.is_typeof(origin) for dt in self.ctypes):
                self.wrong_argument()
        else:
            if not any(map(partial(_type_checker, origin), self.rtypes)):
                self.wrong_argument()
        return origin

class ConverterContainer(Converter, CTConverter):
    def __init_subclass__(subcls) -> None:
        orig_convert = subcls.convert
        orig_cconvert = subcls.cconvert
        def new_convert(self: ConverterContainer, origin: acacia.AcaciaExpr):
            if not all(isinstance(c, Converter) for c in self.__converters):
                raise ChopError(
                    f'{self!r} is used as runtime converter but not all '
                    'subconverters implement Converter'
                )
            return orig_convert(self, origin)
        def new_cconvert(self: ConverterContainer, origin: acaciact.CTObj):
            if not all(isinstance(c, CTConverter) for c in self.__converters):
                raise ChopError(
                    f'{self!r} is used as compile time converter '
                    'but not all subconverters implement CTConverter'
                )
            return orig_cconvert(self, origin)
        subcls.convert = new_convert
        subcls.cconvert = new_cconvert

    def __init__(self, *converters: _CONVERTER_T):
        self.__converters = converters

    def crepr(self) -> str:
        return self.get_show_name()

def _convname(converter: _CONVERTER_T) -> str:
    if isinstance(converter, Converter):
        return converter.get_show_name()
    assert isinstance(converter, CTConverter)
    return converter.crepr()

def _exprrepr(expr: _EXPR_T) -> str:
    if isinstance(expr, acacia.AcaciaExpr):
        return str(expr.data_type)
    assert isinstance(expr, (acaciact.CTObj, acaciact.CTObjPtr))
    return expr.cdata_type.name

class LiteralInt(UTyped):
    """Accepts an integer literal and converts it to Python `int`.
    Default value should also be given as Python `int`.
    """
    def __init__(self):
        super().__init__(objects.IntDataType, objects.integer.ctdt_int)

    def get_show_name(self) -> str:
        return localize("axe.converter.std.literalint")

    def uconvert(self, origin) -> int:
        origin = super().uconvert(origin)
        if isinstance(origin, objects.IntLiteral):
            return origin.value
        self.wrong_argument()

class LiteralFloat(UMultityped):
    """Accepts float and integer literal and converts it to Python
    `float`. Default value should also be given as Python `float`.
    """
    def __init__(self):
        super().__init__(
            (objects.IntDataType, objects.FloatDataType),
            (objects.integer.ctdt_int, objects.float_.ctdt_float)
        )

    def get_show_name(self) -> str:
        return objects.float_.ctdt_float.name

    def uconvert(self, origin) -> float:
        origin = super().uconvert(origin)
        if isinstance(origin, objects.IntLiteral):
            return float(origin.value)
        elif isinstance(origin, objects.Float):
            return origin.value
        self.wrong_argument()

class LiteralString(UTyped):
    """Accepts a string literal and converts it to Python `str`.
    Default value should also be given as Python `str`.
    """
    def __init__(self):
        super().__init__(objects.StringDataType, objects.string.ctdt_string)

    def uconvert(self, origin) -> str:
        origin = super().uconvert(origin)
        assert isinstance(origin, objects.String)
        return origin.value

class LiteralBool(UTyped):
    """Accepts a boolean literal and converts it to Python `bool`.
    Default value should also be given as Python `bool`.
    """
    def __init__(self):
        super().__init__(objects.BoolDataType, objects.boolean.ctdt_bool)

    def get_show_name(self) -> str:
        return localize("axe.converter.std.literalbool")

    def uconvert(self, origin) -> bool:
        origin = super().uconvert(origin)
        if isinstance(origin, objects.BoolLiteral):
            return origin.value
        self.wrong_argument()

class Nullable(ConverterContainer):
    """Accepts "None" value and convert it to Python "None".
    Default value can be set to None.
    """
    def __init__(self, converter: _CONVERTER_T):
        super().__init__(converter)
        self.converter = converter

    def get_show_name(self) -> str:
        return (localize("axe.converter.std.nullable")
                % _convname(self.converter))

    def convert(self, origin: acacia.AcaciaExpr):
        if origin.data_type.matches_cls(objects.NoneDataType):
            return None
        return self.converter.convert(origin)

    def cconvert(self, origin: acaciact.CTObj):
        if isinstance(origin, objects.NoneLiteral):
            return None
        return self.converter.cconvert(origin)

class AnyOf(ConverterContainer):
    """Accepts arguments of several kinds."""
    def __init__(self, *converters: _CONVERTER_T):
        super().__init__(*converters)
        if not converters:
            raise ChopError("at least 1 converter needs to be specified")
        self.converters: List[_CONVERTER_T] = []
        for converter in converters:
            if isinstance(converter, AnyOf):
                self.converters.extend(converter.converters)
            else:
                self.converters.append(converter)

    def get_show_name(self) -> str:
        return " / ".join(_convname(c) for c in self.converters)

    def convert(self, origin: acacia.AcaciaExpr):
        for converter in self.converters:
            try:
                res = converter.convert(origin)
            except _WrongArgTypeError:
                pass
            else:
                return res
        else:
            self.wrong_argument()

    def cconvert(self, origin: acaciact.CTObj):
        for converter in self.converters:
            try:
                res = converter.cconvert(origin)
            except _WrongArgTypeError:
                pass
            else:
                return res
        else:
            self.wrong_argument()

class Iterator(Converter):
    """Accepts an Acacia iterable and converts it to Python list."""
    def get_show_name(self) -> str:
        return localize("axe.converter.std.iterator")

    def convert(self, origin: acacia.AcaciaExpr) -> "acacia.ITERLIST_T":
        try:
            res = origin.iterate()
        except InvalidOpError:
            self.wrong_argument()
        else:
            return res

class CTIterator(CTConverter):
    def crepr(self) -> str:
        return localize("axe.converter.std.ctiterator")

    def cconvert(self, origin: acaciact.CTExpr) -> List[acaciact.CTExpr]:
        try:
            res = abs(origin).citerate()
        except InvalidOpError:
            self.wrong_argument()
        else:
            return res

class Selector(Multityped):
    """Accepts entity or Engroup and convert it to `MCSelector`."""
    def __init__(self):
        super().__init__((objects.EntityDataType, objects.EGroupDataType))

    def convert(self, origin: acacia.AcaciaExpr) -> "MCSelector":
        # Both `_EntityBase` and `EntityGroup` define `get_selector`.
        origin = super().convert(origin)
        return origin.get_selector()

class LiteralIntEnum(LiteralInt):
    """Accepts several specific integer literals values and converts
    input to Python `int`.
    """
    def __init__(self, *accepts: int):
        super().__init__()
        self.accepts = accepts

    def get_show_name(self) -> str:
        return (
            localize("axe.converter.std.literalintenum")
            .format(origin=super().get_show_name(),
                    list=", ".join(map(str, self.accepts)))
        )

    def uconvert(self, origin) -> int:
        origin_int = super().uconvert(origin)
        if origin_int not in self.accepts:
            self.wrong_argument()
        return origin_int

class LiteralStringEnum(LiteralString):
    """Accepts several specific string literals values and converts
    input to Python `str`.
    """
    def __init__(self, *accepts: str):
        super().__init__()
        self.accepts = accepts

    def get_show_name(self) -> str:
        return (
            localize("axe.converter.std.literalstringenum")
            .format(origin=super().get_show_name(),
                    list=", ".join(map(repr, self.accepts)))
        )

    def uconvert(self, origin) -> str:
        origin_str = super().uconvert(origin)
        if origin_str not in self.accepts:
            self.wrong_argument()
        return origin_str

class ListOf(Typed):
    """Accepts a list of specified data type and converts it to Python
    `list`.
    """
    def __init__(self, converter: Converter):
        super().__init__(objects.ListDataType)
        self.converter = converter

    def get_show_name(self) -> str:
        return (localize("axe.converter.std.listof")
                % self.converter.get_show_name())

    def convert(self, origin: acacia.AcaciaExpr) -> list:
        origin = super().convert(origin)
        assert isinstance(origin, objects.AcaciaList)
        return list(map(self.converter.convert, origin.items))

class CTListOf(CTTyped):
    def __init__(self, converter: CTConverter):
        super().__init__(objects.list_.ctdt_constlist)
        self.converter = converter

    def get_show_name(self) -> str:
        return (localize("axe.converter.std.ctlistof")
                % self.converter.crepr())

    def cconvert(self, origin: acaciact.CTExpr) -> list:
        origin = abs(super().cconvert(origin))
        assert isinstance(origin, objects.CTConstList)
        return list(map(self.converter.cconvert, origin.ptrs))

class MapOf(Typed):
    """Accepts a map of specified data type as key and value and
    converts it to Python `dict`.
    """
    def __init__(self, key: Converter, value: Converter):
        super().__init__(objects.MapDataType)
        self.key = key
        self.value = value

    def get_show_name(self) -> str:
        k = self.key.get_show_name()
        v = self.value.get_show_name()
        return localize("axe.converter.std.mapof") % f"({k}: {v})"

    def convert(self, origin: acacia.AcaciaExpr) -> dict:
        origin = super().convert(origin)
        assert isinstance(origin, objects.Map)
        return {
            self.key.convert(key): self.value.convert(value)
            for key, value in origin.items()
        }

class CTMapOf(CTTyped):
    def __init__(self, key: CTConverter, value: CTConverter):
        super().__init__(objects.map_.ctdt_constmap)
        self.key = key
        self.value = value

    def get_show_name(self) -> str:
        k = self.key.crepr()
        v = self.value.crepr()
        return localize("axe.converter.std.ctmapof") % f"({k}: {v})"

    def cconvert(self, origin: acacia.AcaciaExpr) -> dict:
        origin = abs(super().cconvert(origin))
        assert isinstance(origin, objects.CTConstMap)
        return {
            self.key.cconvert(key): self.value.cconvert(value)
            for key, value in origin.items()
        }

class PlayerSelector(Selector):
    """Accepts an entity or Engroup with player type and converts it
    to `MCSelector`.
    """
    def get_show_name(self) -> str:
        return (localize("axe.converter.std.playerselector")
                % super().get_show_name())

    def convert(self, origin: acacia.AcaciaExpr) -> "MCSelector":
        selector = super().convert(origin)
        try:
            selector.player_type()
        except ValueError:
            self.wrong_argument()
        return selector

class RangedLiteralInt(LiteralInt):
    """Accepts a literal integer between given range and converts it to
    Python `int`.
    """
    def __init__(self, min_: Optional[int], max_: Optional[int]):
        super().__init__()
        if min_ is None:
            min_ = -float("inf")
        if max_ is None:
            max_ = float("inf")
        self.min = min_
        self.max = max_

    def get_show_name(self) -> str:
        return super().get_show_name() + " (%s ~ %s)" % (self.min, self.max)

    def uconvert(self, origin) -> int:
        num = super().uconvert(origin)
        if not self.min <= num <= self.max:
            self.wrong_argument()
        return num

class Callable(Converter):
    """Accepts a callable Acacia expression."""
    def get_show_name(self) -> str:
        return localize("axe.converter.std.callable")

    def convert(self, origin: acacia.AcaciaExpr) -> acacia.AcaciaCallable:
        if not isinstance(origin, acacia.AcaciaCallable):
            self.wrong_argument()
        return origin

class PosXZ(LiteralFloat):
    """
    Accepts a literal float or int and converts it to Python `float`.
    If the value is an integer, it gets `0.5` added to it.
    This is used for x and z axis in absolute position, as Minecraft
    also does this (move the position to block center).
    """
    def uconvert(self, origin) -> float:
        res = super().uconvert(origin)
        if isinstance(origin, objects.IntLiteral):
            res += 0.5
        return res

class CTReference(CTConverter):
    def crepr(self) -> str:
        return localize("axe.converter.std.ctreference")

    def cconvert(self, origin: acaciact.CTExpr):
        if not isinstance(origin, acaciact.CTObjPtr):
            self.wrong_argument()
        return origin

### Parser

def _check_repeat(names: List[str], renames: List[_RENAME_T]):
    got = set()
    for name in names:
        if name in got:
            raise ChopError(f"repeated argument {name!r}")
        got.add(name)
    got = set()
    for name in renames:
        if not isinstance(name, str):
            continue
        if name in got:
            raise ChopError(f"repeated argument rename {name!r}")
        got.add(name)

def _call_impl(implementation, all_args: List[str], *args, **kwds):
    try:
        return implementation(*args, **kwds)
    except ArgumentError as err:
        if err.arg not in all_args:
            # Make sure `err.arg` is a valid argument name
            raise ChopError(f"unknown argument {err.arg!r}")
        else:
            raise AcaciaError(ErrorType.INVALID_BIN_FUNC_ARG,
                              arg=err.arg, message=err.message)

def _get_convert(origin: _EXPR_T, converter: _CONVERTER_T):
    if isinstance(origin, acacia.AcaciaExpr):
        if isinstance(converter, Converter):
            convert = lambda: converter.convert(origin)
        elif isinstance(origin, acacia.ConstExpr):
            convert = lambda: converter.cconvert(origin.to_ctexpr())
        else:
            raise _PreconvertError(_PE_NOT_CONST)
    else:
        if isinstance(converter, CTConverter):
            convert = lambda: converter.cconvert(origin)
        else:
            try:
                origin_rt = abs(origin).to_rt()
            except InvalidOpError:
                raise _PreconvertError(_PE_NOT_RT)
            convert = lambda: converter.convert(origin_rt)
    return convert

class _Chopper:
    def __init__(self, building: _BuildingParser):
        self.pos_only: List[_Argument] = []
        self.pos_n_kw: List[_Argument] = []
        self.kw_only: List[_Argument] = []
        self.args: Optional[_ArgumentList] = None
        self.kwds: Optional[_ArgumentList] = None
        self.implementation = building.get_target()
        # Parse builder stack
        got_slash = False
        got_star = False
        got_kwds = False
        got_default = False
        before_slash = []
        for type_, definition in building:
            if type_ == _BP_SLASH:
                if got_slash or got_star or got_kwds:
                    raise ChopError("only arguments can go before axe.slash")
                got_slash = True
            elif type_ == _BP_STAR or type_ == _BP_STAR_ARG:
                if got_star or got_kwds:
                    raise ChopError("only arguments or axe.slash can go before"
                                    " axe.star or axe.star_arg")
                if type_ == _BP_STAR_ARG:
                    self.args = definition
                got_star = True
            elif type_ == _BP_KWDS:
                if got_kwds:
                    raise ChopError("multiple axe.kwds")
                self.kwds = definition
                got_kwds = True
            else:
                assert type_ == _BP_ARG and isinstance(definition, _Argument)
                if got_kwds:
                    raise ChopError("argument after axe.kwds")
                if definition.has_default():
                    got_default = True
                else:
                    if got_default:
                        raise ChopError(
                            "non-default argument follows default argument"
                        )
                if got_star:
                    self.kw_only.append(definition)
                else:
                    if got_slash:
                        self.pos_n_kw.append(definition)
                    else:
                        before_slash.append(definition)
        if got_slash:
            self.pos_only.extend(before_slash)
        else:
            self.pos_n_kw.extend(before_slash)

        names, renames = [], []
        for arg_def in chain(self.pos_only, self.pos_n_kw, self.kw_only):
            names.append(arg_def.name)
            renames.append(arg_def.rename)
        if self.args:
            names.append(self.args.name)
            renames.append(self.args.rename)
        if self.kwds:
            names.append(self.kwds.name)
            renames.append(self.kwds.rename)
        _check_repeat(names, renames)

        self.MAX_POS_ARG = len(self.pos_only) + len(self.pos_n_kw)
        self.pos_only_names = [arg_def.name for arg_def in self.pos_only]
        self.kw_name2def = {arg_def.name: arg_def
                            for arg_def in chain(self.kw_only, self.pos_n_kw)}

    def _convert(self, origin: _EXPR_T, converter: _CONVERTER_T, arg: str):
        try:
            convert = _get_convert(origin, converter)
        except _PreconvertError as err:
            raise AcaciaError(ErrorType.ANY, message=err.show(arg))
        try:
            return convert()
        except _WrongArgTypeError:
            raise AcaciaError(
                ErrorType.WRONG_ARG_TYPE,
                expect=_convname(converter), arg=arg,
                got=_exprrepr(origin)
            )

    def __call__(self, compiler: "Compiler", args: List[_EXPR_T],
                 kwds: Dict[str, _EXPR_T]):
        res: Dict[str, Any] = {}
        res_positional: List[Any] = []
        arg_got: List[str] = []
        def _emit(arg: Union[_Argument, _ArgumentList], value: Any):
            if arg.rename is POSITIONAL:
                res_positional.append(value)
            else:
                res[arg.rename] = value
            arg_got.append(arg.name)
        # Positional arguments
        if len(args) > self.MAX_POS_ARG:
            if self.args is None:
                raise AcaciaError(ErrorType.TOO_MANY_ARGS)
            else:
                vargs = [
                    self._convert(
                        arg, self.args.converter,
                        "#%d(*%s)" % (i, self.args.name)
                    )
                    for i, arg in enumerate(
                        args[self.MAX_POS_ARG:], start=self.MAX_POS_ARG + 1
                    )
                ]
                _emit(self.args, vargs)
        for arg_def, arg in zip(chain(self.pos_only, self.pos_n_kw), args):
            _emit(arg_def, self._convert(arg, arg_def.converter, arg_def.name))
        # Keyword arguments
        extra_kwds = {}
        for arg_name, arg in kwds.items():
            if arg_name in self.pos_only_names:
                raise AcaciaError(
                    ErrorType.ANY,
                    message=localize("axe.chopper.posaskwd") % arg_name
                )
            if arg_name not in self.kw_name2def:
                if self.kwds is None:
                    raise AcaciaError(ErrorType.UNEXPECTED_KEYWORD_ARG,
                                      arg=arg_name)
                else:
                    extra_kwds[arg_name] = arg
                    continue
            if arg_name in arg_got:
                raise AcaciaError(ErrorType.ARG_MULTIPLE_VALUES, arg=arg_name)
            arg_def = self.kw_name2def[arg_name]
            _emit(arg_def, self._convert(arg, arg_def.converter, arg_name))
        if extra_kwds:
            assert self.kwds
            vkwds = {
                arg_name: self._convert(
                    arg, self.kwds.converter,
                    "%s(**%s)" % (arg_name, self.kwds.name)
                )
                for arg_name, arg in extra_kwds.items()
            }
            _emit(self.kwds, vkwds)
        # Check for missing arguments and fix with default values
        for arg_def in chain(self.pos_only, self.pos_n_kw, self.kw_only):
            if arg_def.name not in arg_got:
                try:
                    default = arg_def.get_default()
                except ValueError:
                    raise AcaciaError(ErrorType.MISSING_ARG, arg=arg_def.name)
                else:
                    _emit(arg_def, default)
        if self.args and self.args.name not in arg_got:
            _emit(self.args, [])
        if self.kwds and self.kwds.name not in arg_got:
            _emit(self.kwds, {})
        return _call_impl(self.implementation, arg_got,
                          compiler, *res_positional, **res)

def _create_signature(arg_defs: List[_Argument]) -> str:
    return "(%s)" % ", ".join(
        "%s: %s" % (
            arg_def.name,
            _convname(arg_def.converter)
        )
        for arg_def in arg_defs
    )

class _OverloadImplWrapper:
    def __init__(self, method: PyCallable, building: _BuildingParser,
                 version: Optional["VersionRequirement"]):
        self.method = method
        self.building = building
        self.version = version

    def __call__(self, *args: Any, **kwds: Any) -> Any:
        return self.method(*args, **kwds)

class _OverloadMethod(classmethod):
    def __init__(self, building: _BuildingParser,
                 version: Optional["VersionRequirement"] = None):
        super().__init__(building.get_target())
        self.building = building
        self.version = version

    def __get__(self, instance, owner: Optional[type] = None) -> PyCallable:
        return _OverloadImplWrapper(
            super().__get__(instance, owner), self.building, self.version
        )

### Parser Interface

@_parser_component
def chop(building: _BuildingParser):
    """Use a Python-style argument parser for decorated function.
    Example:
    >>> @chop
    ... @arg("foo", Nullable(Typed(BoolDataType)))
    ... @arg("bar", LiteralInt(), default=11)
    ... def f(compiler, foo, bar):
    ...     # implement this binary function here
    ...     print(foo, bar)
    >>> func = BinaryFunction(f)
    >>> func.call(
    ...     [], {"foo": BoolLiteral(True), "bar": IntLiteral(2)}, compiler
    ... )
    <BoolLiteral object at ...> 2
    >>> func.call([NoneLiteral()], {}, compiler)
    None 11
    """
    return _Chopper(building)

class OverloadChopped(type):
    """Implement a binary function with overload-style argument parser.
    Example:
    >>> class Foo(metaclass=OverloadChopped):
    ...     @overload
    ...     @arg("a", BoolDataType)
    ...     @arg("b", LiteralInt())
    ...     def f1(cls, compiler, a, b):
    ...         print("f1: ", a, b)
    ...     @overload
    ...     @arg("a", BoolDataType)
    ...     def f2(cls, compiler, a):
    ...         print("f2: ", a)
    ...         return cls.f1(compiler, a, b=10)
    >>> func = BinaryFunction(Foo)
    >>> func.call([BoolLiteral(True)], {}, compiler)
    f2: <BoolLiteral object at ...>
    f1: <BoolLiteral object at ...> 10
    >>> func.call([BoolLiteral(False), IntLiteral(5)], {}, compiler)
    f1: <BoolLiteral object at ...> 5
    """

    # For IDE hint only:
    __overloads: List[Tuple[_OverloadImplWrapper, List[_Argument]]]

    def __new__(meta_cls, cls_name, bases, attributes):
        cls = type.__new__(meta_cls, cls_name, bases, attributes)
        # Parse the class
        impls = [attr for _, attr in inspect.getmembers(cls)
                 if isinstance(attr, _OverloadImplWrapper)]
        cls.__overloads = []
        for impl in impls:
            building = impl.building
            if any(type_ != _BP_ARG for type_, _ in building):
                raise ChopError("only normal arguments are allowed in"
                                " overload definitions")
            arg_defs: List[_Argument] = [arg_def for _, arg_def in building]
            if any(arg_def.has_default() for arg_def in arg_defs):
                raise ChopError("overload arguments can't have default values")
            cls.__overloads.append((impl, arg_defs))
            _check_repeat([arg_def.name for arg_def in arg_defs],
                          [arg_def.rename for arg_def in arg_defs])
        return cls

    @staticmethod
    def _format_version(version: Optional["VersionRequirement"]):
        if version is None:
            return ""
        else:
            return " [MC %s]" % version.to_str()

    def __call__(self, compiler: "Compiler", args: List[_EXPR_T],
                 kwds: Dict[str, _EXPR_T]):
        if kwds:
            raise AcaciaError(
                ErrorType.ANY, message=localize("axe.overload.kwd")
            )
        L = len(args)
        for implementation, arg_defs in self.__overloads:
            if len(arg_defs) != L:
                continue
            version = implementation.version
            if version and not version.validate(compiler.cfg.mc_version):
                continue
            res = {}
            for arg_def, arg in zip(arg_defs, args):
                try:
                    convert = _get_convert(arg, arg_def.converter)
                except _PreconvertError:
                    break
                try:
                    converted = convert()
                except _WrongArgTypeError:
                    break
                res[arg_def.rename] = converted
            else:
                return _call_impl(
                    implementation,
                    [arg_def.name for arg_def in arg_defs],
                    compiler, **res
                )
        else:
            raise AcaciaError(
                ErrorType.ANY,
                message=localize("axe.overload.nomatch").format(
                    got="(%s)" % ", ".join(_exprrepr(arg) for arg in args),
                    expected=" / ".join(
                        "%s%s" % (
                            _create_signature(arg_defs),
                            self._format_version(impl.version)
                        )
                        for impl, arg_defs in self.__overloads
                    )
                )
            )

@_parser_component
def overload(building: _BuildingParser):
    """Start an overload implementation in a class decorated with
    metaclass `OverloadChopped`. The decorated implementation will be
    made a `classmethod`.
    """
    return _OverloadMethod(building)

def overload_versioned(version: "VersionRequirement"):
    """Return a decorator that is same as @overload, but decorated
    implementation will only be available when compiler.cfg.mc_version
    satifies given requirements.
    """
    @_parser_component
    def _decorator(building: _BuildingParser):
        return _OverloadMethod(building, version)
    return _decorator



"""
Test module for `acaciamc.tools`.
See "test/module.aca".
"""

from acaciamc.tools import axe
from acaciamc.objects import *

@axe.chop
@axe.arg("x", IntDataType, "y")
@axe.slash
@axe.arg("g", EGroupDataType, axe.POSITIONAL)
@axe.arg("y", axe.Typed(StringDataType), "x", default=None)
@axe.star
@axe.arg("z", axe.Nullable(axe.LiteralString()), default="aa")
@axe.kwds("k", axe.AnyOf(axe.LiteralInt(), axe.LiteralString()))
def _foo(compiler, positional_group, **kwds):
    print(f"foo called with {kwds!r} and entity group {positional_group!r}")

class _bar(metaclass=axe.OverloadChopped):
    @axe.overload
    @axe.arg("x", IntDataType)
    @axe.arg("y", axe.LiteralInt())
    def a(cls, compiler, **kwds):
        print("bar.a called:", kwds)

    @axe.overload
    @axe.arg("x", IntDataType)
    def b(cls, compiler, **kwds):
        print("bar.b called")
        return cls.a(compiler, y=1, **kwds)

class _extbar(_bar):
    @axe.overload
    def c(cls, compiler, **kwds):
        print("extbar.c called")
        return cls.b(compiler, x=IntLiteral(30), **kwds)

def acacia_build(compiler):
    return {
        "foo": BinaryFunction(_foo),
        "bar": BinaryFunction(_bar),
        "extbar": BinaryFunction(_extbar)
    }



"""Modules."""

__all__ = ['ModuleDataType', 'BinaryModule', 'AcaciaModule', 'BuiltModule']

from typing import Dict, Optional
import importlib.util

from acaciamc.mccmdgen.expr import *
from acaciamc.mccmdgen.datatype import DefaultDataType
from acaciamc.mccmdgen.symbol import SymbolTable
from acaciamc.mccmdgen.ctexpr import CTDataType
from acaciamc.error import *

class ModuleDataType(DefaultDataType):
    name = 'module'

ctdt_module = CTDataType("module")

class BuiltModule:
    """Object that should be returned by acacia_build to specify a module."""
    def __init__(
        self,
        attributes: Optional[Dict[str, AcaciaExpr]] = None,
        init_cmds: Optional[CMDLIST_T] = None
    ):
        if attributes is None:
            attributes = {}
        if init_cmds is None:
            init_cmds = []
        self.attributes = attributes
        self.init_cmds = init_cmds

class BinaryModule(ConstExprCombined):
    """A binary module that is implemented in Python."""
    cdata_type = ctdt_module

    def __init__(self, path: str):
        """`path` is path to the binary module Python file."""
        super().__init__(ModuleDataType())
        self.path = path
        # get the module from `path`
        self.spec = importlib.util.spec_from_file_location(
            '<acacia module %r>' % path, path
        )
        self.py_module = importlib.util.module_from_spec(self.spec)

    def execute(self, compiler) -> CMDLIST_T:
        """
        Execute the code in the binary module Python file.
        Return the commands run by the module during initialization.
        """
        self.spec.loader.exec_module(self.py_module)
        # Call `acacia_build`
        # Binary modules should define a function named `acacia_build`,
        # which accepts 1 argument `compiler` and should return either
        # a `BuiltModule` or a `dict` containing the attributes of the
        # module (same as `BuiltModule.attributes`).
        res = self.py_module.acacia_build(compiler)
        if isinstance(res, dict):
            res = BuiltModule(res)
        for name, value in res.attributes.items():
            self.attribute_table.set(name, value)
        return res.init_cmds

class AcaciaModule(ConstExprCombined):
    """An Acacia module that is implemented in Acacia."""
    cdata_type = ctdt_module

    def __init__(self, table: SymbolTable):
        super().__init__(ModuleDataType())
        for name in table.all_names():
            if not (name in table.no_export or name.startswith('_')):
                self.attribute_table.set(name, table.get_raw(name))
